from __future__ import annotations

import copy
import pickle
import weakref
from abc import ABCMeta, abstractmethod
from dataclasses import dataclass
from typing import (
    Annotated,
    Any,
    Callable,
    Generic,
    Mapping,
    Optional,
    Sequence,
    TypeVar,
    Union,
)

import pytest
from typing_extensions import Self

from koerce import (
    EMPTY,
    Abstract,
    AbstractMeta,
    Annotable,
    AnnotableMeta,
    Anything,
    As,
    Attribute,
    FrozenDictOf,
    Hashable,
    Immutable,
    Is,
    IsType,
    MatchError,
    Option,
    Parameter,
    Pattern,
    Signature,
    TupleOf,
    annotated,
    argument,
    attribute,
    optional,
    pattern,
    varargs,
    varkwargs,
)


def test_parameter():
    p = Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=int)
    assert p.kind is Parameter.POSITIONAL_OR_KEYWORD
    assert p.format("x") == "x: int"
    assert hash(p) == hash(p)

    p = Parameter(Parameter.POSITIONAL_OR_KEYWORD, default=1)
    assert p.kind is Parameter.POSITIONAL_OR_KEYWORD
    assert p.default_ == 1
    assert p.format("x") == "x=1"
    assert p.pattern == Anything()
    assert hash(p) == hash(p)

    p = Parameter(
        Parameter.POSITIONAL_OR_KEYWORD, typehint=int, default=1, pattern=is_int
    )
    assert p.kind is Parameter.POSITIONAL_OR_KEYWORD
    assert p.default_ == 1
    assert p.typehint is int
    assert p.format("x") == "x: int = 1"
    assert p.pattern == is_int
    assert hash(p) == hash(p)

    p = Parameter(Parameter.VAR_POSITIONAL, typehint=int, pattern=is_int)
    assert p.kind is Parameter.VAR_POSITIONAL
    assert p.typehint is int
    assert p.format("y") == "*y: int"
    assert p.pattern == TupleOf(is_int)
    assert hash(p) == hash(p)

    p = Parameter(Parameter.VAR_KEYWORD, typehint=int, pattern=is_int)
    assert p.kind is Parameter.VAR_KEYWORD
    assert p.typehint is int
    assert p.format("z") == "**z: int"
    assert p.pattern == FrozenDictOf(Anything(), is_int)
    assert hash(p) == hash(p)

    p = Parameter(Parameter.VAR_KEYWORD, typehint=int, pattern=is_int, default={})
    assert p.kind is Parameter.VAR_KEYWORD
    assert p.typehint is int
    assert p.format("z") == "**z: int = {}"
    assert p.pattern == FrozenDictOf(Anything(), is_int)
    assert hash(p) == hash(p)


def test_attribute():
    a = Attribute(int, default=1)
    assert a.pattern == Is(int)
    assert a.default_ == 1
    assert hash(a) == hash(a)

    a = Attribute(dict, default={})
    assert a.pattern == Is(dict)
    assert a.default_ == {}
    assert hash(a) == hash(a)

    a = Attribute(dict[str, float], default={"a": 1.0})
    assert a.pattern == Is(dict[str, float])
    assert a.default_ == {"a": 1.0}
    assert hash(a) == hash(a)


def test_signature_contruction():
    a = Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=int)
    b = Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=str)
    c = Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=int, default=1)
    d = Parameter(Parameter.VAR_POSITIONAL, typehint=int, pattern=is_int)

    sig = Signature({"a": a, "b": b, "c": c, "d": d})
    assert sig.parameters == {"a": a, "b": b, "c": c, "d": d}
    assert sig.return_typehint is EMPTY
    assert sig.return_pattern == Anything()
    assert hash(sig) == hash(sig)


def test_signature_equality_comparison():
    # order of parameters matters
    a = Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=int)
    b = Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=str)
    c = Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=int, default=1)

    sig1 = Signature({"a": a, "b": b, "c": c})
    sig2 = Signature({"a": a, "b": b, "c": c})
    assert sig1 == sig2
    assert hash(sig1) == hash(sig2)

    sig3 = Signature({"a": a, "c": c, "b": b})
    assert sig1 != sig3
    assert hash(sig1) != hash(sig3)


def test_signature_from_callable():
    def func(a: int, b: str, *args, c=1, **kwargs) -> float: ...

    sig = Signature.from_callable(func)
    assert sig.parameters == {
        "a": Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=int),
        "b": Parameter(Parameter.POSITIONAL_OR_KEYWORD, typehint=str),
        "args": Parameter(Parameter.VAR_POSITIONAL),
        "c": Parameter(Parameter.KEYWORD_ONLY, default=1),
        "kwargs": Parameter(Parameter.VAR_KEYWORD),
    }
    assert sig.return_typehint is float


def test_signature_bind_various():
    # with positional or keyword default
    def func(a: int, b: str, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2")
    assert bound == {"a": 1, "b": "2", "c": 1}

    # with variable positional arguments
    def func(a: int, b: str, *args: int, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2", 3, 4)
    assert bound == {"a": 1, "b": "2", "args": (3, 4), "c": 1}

    # with both variadic positional and variadic keyword arguments
    def func(a: int, b: str, *args: int, c=1, **kwargs: int) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2", 3, 4, x=5, y=6)
    assert bound == {
        "a": 1,
        "b": "2",
        "args": (3, 4),
        "c": 1,
        "kwargs": {"x": 5, "y": 6},
    }

    # with positional only arguments
    def func(a: int, b: str, /, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2")
    assert bound == {"a": 1, "b": "2", "c": 1}

    with pytest.raises(TypeError, match="passed as keyword argument"):
        sig(a=1, b="2", c=3)

    # with keyword only arguments
    def func(a: int, b: str, *, c=1) -> float: ...

    sig = Signature.from_callable(func)
    bound = sig(1, "2", c=3)
    assert bound == {"a": 1, "b": "2", "c": 3}

    with pytest.raises(TypeError, match="too many positional arguments"):
        sig(1, "2", 3)

    def func(a, *args, b, z=100, **kwargs): ...

    sig = Signature.from_callable(func)
    bound = sig(10, 20, b=30, c=40, args=50, kwargs=60)
    assert bound == {
        "a": 10,
        "args": (20,),
        "b": 30,
        "z": 100,
        "kwargs": {"c": 40, "args": 50, "kwargs": 60},
    }


def call(func, *args, **kwargs):
    # it also tests the unbind method
    sig = Signature.from_callable(func)
    bound = sig(*args, **kwargs)
    ubargs, ubkwargs = sig.unbind(bound)
    return func(*ubargs, **ubkwargs)


def test_signature_bind_no_arguments():
    def func(): ...

    sig = Signature.from_callable(func)
    assert sig() == {}

    with pytest.raises(TypeError, match="too many positional arguments"):
        sig(1)
    with pytest.raises(TypeError, match="too many positional arguments"):
        sig(1, keyword=2)
    with pytest.raises(TypeError, match="got an unexpected keyword argument 'keyword'"):
        sig(keyword=1)


def test_signature_bind_positional_or_keyword_arguments():
    def func(a, b, c):
        return a, b, c

    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func)
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, 1)
    with pytest.raises(TypeError, match="missing a required argument: 'c'"):
        call(func, 1, 2)
    assert call(func, 1, 2, 3) == (1, 2, 3)

    # one optional argument
    def func(a, b, c=0):
        return a, b, c

    assert call(func, 1, 2, 3) == (1, 2, 3)
    assert call(func, 1, 2) == (1, 2, 0)

    # two optional arguments
    def func(a, b=0, c=0):
        return a, b, c

    assert call(func, 1, 2, 3) == (1, 2, 3)
    assert call(func, 1, 2) == (1, 2, 0)
    assert call(func, 1) == (1, 0, 0)

    # three optional arguments
    def func(a=0, b=0, c=0):
        return a, b, c

    assert call(func, 1, 2, 3) == (1, 2, 3)
    assert call(func, 1, 2) == (1, 2, 0)
    assert call(func, 1) == (1, 0, 0)
    assert call(func) == (0, 0, 0)


def test_signature_bind_varargs():
    def func(*args):
        return args

    assert call(func) == ()
    assert call(func, 1) == (1,)
    assert call(func, 1, 2) == (1, 2)
    assert call(func, 1, 2, 3) == (1, 2, 3)

    def func(a, b, c=3, *args):
        return a, b, c, args

    assert call(func, 1, 2) == (1, 2, 3, ())
    assert call(func, 1, 2, 3) == (1, 2, 3, ())
    assert call(func, 1, 2, 3, 4) == (1, 2, 3, (4,))
    assert call(func, 1, 2, 3, 4, 5) == (1, 2, 3, (4, 5))
    assert call(func, 1, 2, 4) == (1, 2, 4, ())
    assert call(func, a=1, b=2, c=3) == (1, 2, 3, ())
    assert call(func, c=3, a=1, b=2) == (1, 2, 3, ())

    with pytest.raises(TypeError, match="multiple values for argument 'c'"):
        call(func, 1, 2, 3, c=4)

    def func(a, *args):
        return a, args

    with pytest.raises(TypeError, match="got an unexpected keyword argument 'args'"):
        call(func, a=0, args=1)

    def func(*args, **kwargs):
        return args, kwargs

    assert call(func, args=1) == ((), {"args": 1})

    sig = Signature.from_callable(func)
    ba = sig(args=1)
    assert ba == {"args": (), "kwargs": {"args": 1}}


def test_signature_bind_varkwargs():
    def func(**kwargs):
        return kwargs

    assert call(func) == {}
    assert call(func, foo="bar") == {"foo": "bar"}
    assert call(func, foo="bar", spam="ham") == {"foo": "bar", "spam": "ham"}

    def func(a, b, c=3, **kwargs):
        return a, b, c, kwargs

    assert call(func, 1, 2) == (1, 2, 3, {})
    assert call(func, 1, 2, foo="bar") == (1, 2, 3, {"foo": "bar"})
    assert call(func, 1, 2, foo="bar", spam="ham") == (
        1,
        2,
        3,
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, 1, 2, foo="bar", spam="ham", c=4) == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, 1, 2, c=4, foo="bar", spam="ham") == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, 1, 2, c=4, foo="bar", spam="ham", args=10) == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham", "args": 10},
    )
    assert call(func, b=2, a=1, c=4, foo="bar", spam="ham") == (
        1,
        2,
        4,
        {"foo": "bar", "spam": "ham"},
    )


def test_signature_bind_varargs_and_varkwargs():
    def func(*args, **kwargs):
        return args, kwargs

    assert call(func) == ((), {})
    assert call(func, 1) == ((1,), {})
    assert call(func, 1, 2) == ((1, 2), {})
    assert call(func, foo="bar") == ((), {"foo": "bar"})
    assert call(func, 1, foo="bar") == ((1,), {"foo": "bar"})
    assert call(func, args=10), () == {"args": 10}
    assert call(func, 1, 2, foo="bar") == ((1, 2), {"foo": "bar"})
    assert call(func, 1, 2, foo="bar", spam="ham") == (
        (1, 2),
        {"foo": "bar", "spam": "ham"},
    )
    assert call(func, foo="bar", spam="ham", args=10) == (
        (),
        {"foo": "bar", "spam": "ham", "args": 10},
    )


def test_signature_bind_positional_only_arguments():
    def func(a, b, /, c=3):
        return a, b, c

    assert call(func, 1, 2) == (1, 2, 3)
    assert call(func, 1, 2, 4) == (1, 2, 4)
    assert call(func, 1, 2, c=4) == (1, 2, 4)
    with pytest.raises(TypeError, match="multiple values for argument 'c'"):
        call(func, 1, 2, 3, c=4)

    def func(a, b=2, /, c=3, *args):
        return a, b, c, args

    assert call(func, 1, 2) == (1, 2, 3, ())
    assert call(func, 1, 2, 4) == (1, 2, 4, ())
    assert call(func, 1, c=3) == (1, 2, 3, ())

    def func(a, b, c=3, /, foo=42, *, bar=50, **kwargs):
        return a, b, c, foo, bar, kwargs

    assert call(func, 1, 2, 4, 5, bar=6) == (1, 2, 4, 5, 6, {})
    assert call(func, 1, 2) == (1, 2, 3, 42, 50, {})
    assert call(func, 1, 2, foo=4, bar=5) == (1, 2, 3, 4, 5, {})
    assert call(func, 1, 2, foo=4, bar=5, c=10) == (1, 2, 3, 4, 5, {"c": 10})
    assert call(func, 1, 2, 30, c=31, foo=4, bar=5) == (1, 2, 30, 4, 5, {"c": 31})
    assert call(func, 1, 2, 30, foo=4, bar=5, c=31) == (1, 2, 30, 4, 5, {"c": 31})
    assert call(func, 1, 2, c=4) == (1, 2, 3, 42, 50, {"c": 4})
    assert call(func, 1, 2, c=4, foo=5) == (1, 2, 3, 5, 50, {"c": 4})

    with pytest.raises(
        TypeError, match="positional only argument 'a' passed as keyword argument"
    ):
        call(func, a=1, b=2)

    def func(a=1, b=2, /):
        return a, b

    with pytest.raises(TypeError, match="got an unexpected keyword argument 'a'"):
        call(func, a=3, b=4)

    def func(a, /, **kwargs):
        return a, kwargs

    assert call(func, "pos-only", bar="keyword") == ("pos-only", {"bar": "keyword"})


def test_signature_bind_keyword_only_arguments():
    def func(*, a, b, c=3):
        return a, b, c

    with pytest.raises(TypeError, match="too many positional arguments"):
        call(func, 1)

    assert call(func, a=1, b=2) == (1, 2, 3)
    assert call(func, a=1, b=2, c=4) == (1, 2, 4)

    def func(a, *, b, c=3, **kwargs):
        return a, b, c, kwargs

    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, 1)

    assert call(func, 1, b=2) == (1, 2, 3, {})
    assert call(func, 1, b=2, c=4) == (1, 2, 4, {})

    def func(*, a, b, c=3, foo=42, **kwargs):
        return a, b, c, foo, kwargs

    assert call(func, a=1, b=2) == (1, 2, 3, 42, {})
    assert call(func, a=1, b=2, foo=4) == (1, 2, 3, 4, {})
    assert call(func, a=1, b=2, foo=4, bar=5) == (1, 2, 3, 4, {"bar": 5})
    assert call(func, a=1, b=2, foo=4, bar=5, c=10) == (1, 2, 10, 4, {"bar": 5})
    assert call(func, a=1, b=2, foo=4, bar=5, c=10, spam=6) == (
        1,
        2,
        10,
        4,
        {"bar": 5, "spam": 6},
    )

    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func, b=2)
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, a=1)

    def func(a, *, b):
        return a, b

    assert call(func, 1, b=2) == (1, 2)
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, 1)
    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func, b=2)
    with pytest.raises(TypeError, match="got an unexpected keyword argument 'c'"):
        call(func, a=1, b=2, c=3)
    with pytest.raises(TypeError, match="too many positional arguments"):
        call(func, 1, 2)
    with pytest.raises(TypeError, match="too many positional arguments"):
        call(func, 1, 2, c=3)

    def func(a, *, b, **kwargs):
        return a, b, kwargs

    assert call(func, 1, b=2) == (1, 2, {})
    assert call(func, 1, b=2, c=3) == (1, 2, {"c": 3})
    assert call(func, 1, b=2, c=3, d=4) == (1, 2, {"c": 3, "d": 4})
    assert call(func, a=1, b=2) == (1, 2, {})
    assert call(func, c=3, a=1, b=2) == (1, 2, {"c": 3})
    with pytest.raises(TypeError, match="missing a required argument: 'b'"):
        call(func, a=1)
    with pytest.raises(TypeError, match="missing a required argument: 'a'"):
        call(func, c=3, b=2)


def test_signature_bind_with_arg_named_self():
    def test(a, self, b):
        pass

    sig = Signature.from_callable(test)
    ba = sig(1, 2, 3)
    args, _ = sig.unbind(ba)
    assert args == (1, 2, 3)
    ba = sig(1, self=2, b=3)
    args, _ = sig.unbind(ba)
    assert args == (1, 2, 3)


def test_signature_unbind_from_callable():
    def test(a: int, b: int, c: int = 1): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3)

    assert bound == {"a": 2, "b": 3, "c": 1}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 1)
    assert kwargs == {}


def test_signature_unbind_from_callable_with_varargs():
    def test(a: int, b: int, *args: int): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3)

    assert bound == {"a": 2, "b": 3, "args": ()}
    args, kwargs = sig.unbind(bound)
    assert args == (2, 3)
    assert kwargs == {}

    bound = sig(2, 3, 4, 5)
    assert bound == {"a": 2, "b": 3, "args": (4, 5)}
    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 4, 5)
    assert kwargs == {}


def test_signature_unbind_from_callable_with_positional_only_arguments():
    def test(a: int, b: int, /, c: int = 1): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3)
    assert bound == {"a": 2, "b": 3, "c": 1}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 1)
    assert kwargs == {}

    bound = sig(2, 3, 4)
    assert bound == {"a": 2, "b": 3, "c": 4}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3, 4)
    assert kwargs == {}


def test_signature_unbind_from_callable_with_keyword_only_arguments():
    def test(a: int, b: int, *, c: float, d: float = 0.0): ...

    sig = Signature.from_callable(test)
    bound = sig(2, 3, c=4.0)
    assert bound == {"a": 2, "b": 3, "c": 4.0, "d": 0.0}

    args, kwargs = sig.unbind(bound)
    assert args == (2, 3)
    assert kwargs == {"c": 4.0, "d": 0.0}


def test_signature_unbind():
    def func(a, b, c=1): ...

    sig = Signature.from_callable(func)
    bound = sig(1, 2)

    assert bound == {"a": 1, "b": 2, "c": 1}

    args, kwargs = sig.unbind(bound)
    assert args == (1, 2, 1)
    assert kwargs == {}


@pytest.mark.parametrize("d", [(), (5, 6, 7)])
def test_signature_unbind_with_empty_variadic(d):
    def func(a, b, c, *args, e=None):
        return a, b, c, args, e

    sig = Signature.from_callable(func)
    bound = sig(1, 2, 3, *d, e=4)
    assert bound == {"a": 1, "b": 2, "c": 3, "args": d, "e": 4}

    args, kwargs = sig.unbind(bound)
    assert args == (1, 2, 3, *d)
    assert kwargs == {"e": 4}


def test_signature_merge():
    def f1(a, b, c=1): ...

    def f2(d, e, f=2): ...

    def f3(d, a=1, **kwargs): ...

    sig1 = Signature.from_callable(f1)
    sig2 = Signature.from_callable(f2)
    sig3 = Signature.from_callable(f3)

    sig12 = Signature.merge([sig1, sig2])
    assert sig12.parameters == {
        "a": Parameter(Parameter.POSITIONAL_OR_KEYWORD),
        "b": Parameter(Parameter.POSITIONAL_OR_KEYWORD),
        "d": Parameter(Parameter.POSITIONAL_OR_KEYWORD),
        "e": Parameter(Parameter.POSITIONAL_OR_KEYWORD),
        "c": Parameter(Parameter.POSITIONAL_OR_KEYWORD, default=1),
        "f": Parameter(Parameter.POSITIONAL_OR_KEYWORD, default=2),
    }
    assert tuple(sig12.parameters.keys()) == ("a", "b", "d", "e", "c", "f")

    sig13 = Signature.merge([sig1, sig3])
    assert sig13.parameters == {
        "b": Parameter(Parameter.POSITIONAL_OR_KEYWORD),
        "d": Parameter(Parameter.POSITIONAL_OR_KEYWORD),
        "a": Parameter(Parameter.POSITIONAL_OR_KEYWORD, default=1),
        "c": Parameter(Parameter.POSITIONAL_OR_KEYWORD, default=1),
        "kwargs": Parameter(Parameter.VAR_KEYWORD),
    }
    assert tuple(sig13.parameters.keys()) == ("b", "d", "a", "c", "kwargs")


def test_annotated_function():
    @annotated(a=IsType(int), b=IsType(int), c=IsType(int))
    def test(a, b, c=1):
        return a + b + c

    assert test(2, 3) == 6
    assert test(2, 3, 4) == 9
    assert test(2, 3, c=4) == 9
    assert test(a=2, b=3, c=4) == 9

    with pytest.raises(MatchError):
        test(2, 3, c="4")

    @annotated(a=IsType(int))
    def test(a, b, c=1):
        return (a, b, c)

    assert test(2, "3") == (2, "3", 1)


def test_annotated_function_with_type_annotations():
    @annotated()
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    assert test(2, 3) == 6

    @annotated
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    assert test(2, 3) == 6

    @annotated
    def test(a: int, b, c=1):
        return (a, b, c)

    assert test(2, 3, "4") == (2, 3, "4")


def test_annotated_function_with_return_type_annotation():
    @annotated
    def test_ok(a: int, b: int, c: int = 1) -> int:
        return a + b + c

    @annotated
    def test_wrong(a: int, b: int, c: int = 1) -> int:
        return "invalid result"

    assert test_ok(2, 3) == 6
    with pytest.raises(MatchError):
        test_wrong(2, 3)


def test_annotated_function_with_keyword_overrides():
    @annotated(b=IsType(float))
    def test(a: int, b: int, c: int = 1):
        return a + b + c

    with pytest.raises(MatchError):
        test(2, 3)

    assert test(2, 3.0) == 6.0


def test_annotated_function_with_list_overrides():
    with pytest.raises(MatchError):

        @annotated([IsType(int), IsType(int), IsType(str)])
        def test(a: int, b: int, c: int = 1):
            return a + b + c

    @annotated([IsType(int), IsType(int), IsType(float)])
    def test(a: int, b: int, c: int = 1.0):
        return a + b + c

    assert test(2, 3) == 6.0
    assert isinstance(test(2, 3), float)
    with pytest.raises(MatchError):
        test(2, 3, 4)


def test_annotated_function_with_list_overrides_and_return_override():
    with pytest.raises(MatchError):

        @annotated([IsType(int), IsType(int), IsType(float)], IsType(float))
        def test(a: int, b: int, c: int = 1):
            return a + b + c

    @annotated([IsType(int), IsType(int), IsType(float)], IsType(float))
    def test(a: int, b: int, c: int = 1.1):
        return a + b + c

    assert test(2, 3) == 6.1
    with pytest.raises(MatchError):
        test(2, 3, 4)

    assert test(2, 3, 4.0) == 9.0


@pattern
def short_str(x, **context):
    if len(x) > 3:
        return x
    else:
        raise ValueError("string is too short")


@pattern
def endswith_d(x, **context):
    if x.endswith("d"):
        return x
    else:
        raise ValueError("string does not end with 'd'")


def test_annotated_function_with_complex_type_annotations():
    @annotated
    def test(a: Annotated[str, short_str, endswith_d], b: Union[int, float]):
        return a, b

    assert test("abcd", 1) == ("abcd", 1)
    assert test("---d", 1.0) == ("---d", 1.0)

    with pytest.raises(MatchError):
        test("---c", 1)
    with pytest.raises(MatchError):
        test("123", 1)
    with pytest.raises(MatchError):
        test("abcd", "qweqwe")


def test_annotated_function_without_annotations():
    @annotated
    def test(a, b, c):
        return a, b, c

    assert test(1, 2, 3) == (1, 2, 3)
    assert list(test.__signature__.parameters.keys()) == ["a", "b", "c"]


def test_annotated_function_without_decoration():
    def test(a, b, c):
        return a + b + c

    func = annotated(test)
    with pytest.raises(TypeError):
        func(1, 2)

    assert func(1, 2, c=3) == 6


def test_annotated_function_with_varargs():
    @annotated
    def test(a: float, b: float, *args: int):
        return sum((a, b) + args)

    assert test(1.0, 2.0, 3, 4) == 10.0
    assert test(1.0, 2.0, 3, 4, 5) == 15.0
    assert test(1.0, 2.0, 3, 4, 5, 6.0) == 21.0


def test_annotated_function_with_varkwargs():
    @annotated
    def test(a: float, b: float, **kwargs: int):
        return sum((a, b) + tuple(kwargs.values()))

    assert test(1.0, 2.0, c=3, d=4) == 10.0
    assert test(1.0, 2.0, c=3, d=4, e=5) == 15.0
    assert test(1.0, 2.0, c=3, d=4, e=5, f=6.0) == 21.0


# def test_multiple_validation_failures():
#     @annotated
#     def test(a: float, b: float, *args: int, **kwargs: int): ...

#     with pytest.raises(ValidationError) as excinfo:
#         test(1.0, 2.0, 3.0, 4, c=5.0, d=6)

#     assert len(excinfo.value.errors) == 2


def test_signature_patterns():
    def func(a: int, b: str) -> str: ...

    sig = Signature.from_callable(func, allow_coercion=False)
    assert sig.parameters["a"].pattern == IsType(int)
    assert sig.parameters["b"].pattern == IsType(str)
    assert sig.return_pattern == IsType(str)

    def func(a: int, b: str, c: str = "0") -> str: ...

    sig = Signature.from_callable(func, allow_coercion=True)
    assert sig.parameters["a"].pattern == As(int)
    assert sig.parameters["b"].pattern == Is(str)
    assert sig.parameters["c"].pattern == Is(str)
    assert sig.return_pattern == IsType(str)

    def func(a: int, b: str, *args): ...

    sig = Signature.from_callable(func)
    assert sig.parameters["a"].pattern == As(int)
    assert sig.parameters["b"].pattern == Is(str)
    assert sig.parameters["args"].pattern == TupleOf(Anything())
    assert sig.return_pattern == Anything()

    def func(a: int, b: str, c: str = "0", *args, **kwargs: int) -> float: ...

    sig = Signature.from_callable(func)
    assert sig.parameters["a"].pattern == As(int)
    assert sig.parameters["b"].pattern == Is(str)
    assert sig.parameters["c"].pattern == Is(str)
    assert sig.parameters["args"].pattern == TupleOf(Anything())
    assert sig.parameters["kwargs"].pattern == FrozenDictOf(Anything(), As(int))
    assert sig.return_pattern == As(float)


def test_annotated_with_class():
    @annotated
    class A:
        a: int
        b: str
        c: float
        d: Optional[int]

        def __init__(self, a, b, c, d=1):
            self.a = a
            self.b = b
            self.c = c
            self.d = d

    with pytest.raises(MatchError):
        A(1, "2", "d")


def test_annotated_with_dataclass():
    @annotated
    @dataclass
    class InventoryItem:
        name: str
        unit_price: float
        quantity_on_hand: int = 10

    @annotated
    @dataclass
    class InventoryItemStrict:
        name: str
        unit_price: Is[float]
        quantity_on_hand: Is[int] = 0

    @annotated
    @dataclass
    class InventoryItemLoose:
        name: str
        unit_price: As[float]
        quantity_on_hand: As[int] = 10

    items = [
        InventoryItem("widget", 3.0, 10),
        InventoryItem("widget", 3.0),
        InventoryItem("widget", "3.0", 10),
        InventoryItem("widget", 3.0, "10"),
        InventoryItemLoose("widget", 3.0, 10),
        InventoryItemLoose("widget", 3.0),
        InventoryItemLoose("widget", "3.0", 10),
        InventoryItemLoose("widget", 3.0, "10"),
    ]
    for item in items:
        assert item.name == "widget"
        assert item.unit_price == 3.0
        assert item.quantity_on_hand == 10

    with pytest.raises(MatchError):
        InventoryItem("widget", 3.0, "10.1")
    with pytest.raises(MatchError):
        InventoryItem("widget", 3.0, 10.1)

    item = InventoryItemStrict("widget", 3.0, 10)
    assert item.name == "widget"
    assert item.unit_price == 3.0
    assert item.quantity_on_hand == 10

    with pytest.raises(MatchError):
        InventoryItemStrict("widget", "3.0", 10)
    with pytest.raises(MatchError):
        InventoryItemStrict("widget", 3.0, "10")


##################################################


is_any = IsType(object)
is_bool = IsType(bool)
is_float = IsType(float)
is_int = IsType(int)
is_str = IsType(str)
is_list = IsType(list)


class Op(Annotable):
    pass


class Value(Op):
    arg = argument(IsType(object))


class StringOp(Value):
    arg = argument(IsType(str))


class BetweenSimple(Annotable):
    value = argument(is_int)
    lower = optional(is_int, default=0)
    upper = optional(is_int, default=None)


class BetweenWithExtra(Annotable):
    extra = attribute(is_int)
    value = argument(is_int)
    lower = optional(is_int, default=0)
    upper = optional(is_int, default=None)


class BetweenWithCalculated(Annotable, immutable=True, hashable=True):
    value = argument(is_int)
    lower = optional(is_int, default=0)
    upper = optional(is_int, default=None)

    @attribute
    def calculated(self):
        return self.value + self.lower


class VariadicArgs(Annotable, immutable=True, hashable=True):
    args = varargs(is_int)


class VariadicKeywords(Annotable, immutable=True, hashable=True):
    kwargs = varkwargs(is_int)


class VariadicArgsAndKeywords(Annotable, immutable=True, hashable=True):
    args = varargs(is_int)
    kwargs = varkwargs(is_int)


T = TypeVar("T", covariant=True)
K = TypeVar("K", covariant=True)
V = TypeVar("V", covariant=True)


class List(Annotable, Generic[T]):
    @classmethod
    def __coerce__(self, values, T=None):
        values = tuple(values)
        if values:
            head, *rest = values
            return ConsList(head, rest)
        else:
            return EmptyList()

    def __eq__(self, other) -> bool:
        if not isinstance(other, List):
            return NotImplemented
        if len(self) != len(other):
            return False
        for a, b in zip(self, other):
            if a != b:
                return False
        return True


# AnnotableMeta doesn't extend ABCMeta, so we need to register the class
# this is due to performance reasons since ABCMeta overrides
# __instancecheck__ and __subclasscheck__ which makes
# issubclass and isinstance way slower
Sequence.register(List)


class EmptyList(List[T]):
    def __getitem__(self, key):
        raise IndexError(key)

    def __len__(self):
        return 0


class ConsList(List[T]):
    head: T
    rest: List[T]

    def __getitem__(self, key):
        if key == 0:
            return self.head
        else:
            return self.rest[key - 1]

    def __len__(self):
        return len(self.rest) + 1


class Map(Annotable, Generic[K, V]):
    @classmethod
    def __coerce__(self, pairs, K=None, V=None):
        pairs = dict(pairs)
        if pairs:
            head_key = next(iter(pairs))
            head_value = pairs.pop(head_key)
            rest = pairs
            return ConsMap((head_key, head_value), rest)
        else:
            return EmptyMap()

    def __eq__(self, other) -> bool:
        if not isinstance(other, Map):
            return NotImplemented
        if len(self) != len(other):
            return False
        for key in self:
            if self[key] != other[key]:
                return False
        return True

    def items(self):
        for key in self:
            yield key, self[key]


# AnnotableMeta doesn't extend ABCMeta, so we need to register the class
# this is due to performance reasons since ABCMeta overrides
# __instancecheck__ and __subclasscheck__ which makes
# issubclass and isinstance way slower
Mapping.register(Map)


class EmptyMap(Map[K, V]):
    def __getitem__(self, key):
        raise KeyError(key)

    def __iter__(self):
        return iter(())

    def __len__(self):
        return 0


class ConsMap(Map[K, V]):
    head: tuple[K, V]
    rest: Map[K, V]

    def __getitem__(self, key):
        if key == self.head[0]:
            return self.head[1]
        else:
            return self.rest[key]

    def __iter__(self):
        yield self.head[0]
        yield from self.rest

    def __len__(self):
        return len(self.rest) + 1


class Integer(int):
    @classmethod
    def __coerce__(cls, value):
        return Integer(value)


class Float(float):
    @classmethod
    def __coerce__(cls, value):
        return Float(value)


class MyExpr(Annotable):
    a: Integer
    b: List[Float]
    c: Map[str, Integer]


class MyInt(int):
    @classmethod
    def __coerce__(cls, value):
        return cls(value)


class MyFloat(float):
    @classmethod
    def __coerce__(cls, value):
        return cls(value)


J = TypeVar("J", bound=MyInt, covariant=True)
F = TypeVar("F", bound=MyFloat, covariant=True)
N = TypeVar("N", bound=Union[MyInt, MyFloat], covariant=True)


class MyValue(Annotable, Generic[J, F]):
    integer: J
    floating: F
    numeric: N


def test_annotable():
    class Between(BetweenSimple):
        pass

    assert not issubclass(type(Between), ABCMeta)
    assert type(Between) is AnnotableMeta

    argnames = ("value", "lower", "upper")
    signature = BetweenSimple.__signature__
    assert isinstance(signature, Signature)
    assert BetweenSimple.__slots__ == argnames

    obj = BetweenSimple(10, lower=2)
    assert obj.value == 10
    assert obj.lower == 2
    assert obj.upper is None
    assert obj.__argnames__ == argnames
    assert obj.__slots__ == ("value", "lower", "upper")
    assert not hasattr(obj, "__dict__")
    assert obj.__module__ == __name__
    assert type(obj).__qualname__ == "BetweenSimple"

    # test that a child without additional arguments doesn't have __dict__
    obj = Between(10, lower=2)
    assert obj.__slots__ == tuple()
    assert not hasattr(obj, "__dict__")
    assert obj.__module__ == __name__
    assert type(obj).__qualname__ == "test_annotable.<locals>.Between"

    copied = copy.copy(obj)
    assert obj == copied
    assert obj is not copied

    # copied = obj.copy()
    # assert obj == copied
    # assert obj is not copied

    # obj2 = Between(10, lower=8)
    # assert obj.copy(lower=8) == obj2


def test_annotable_keeps_annotations_classvar():
    class MyClass(Annotable):
        a: int
        b: str

    assert MyClass.__annotations__ == {"a": "int", "b": "str"}


def test_annotable_with_bound_typevars_properly_coerce_values():
    v = MyValue(1.1, 2.2, 3.3)
    assert isinstance(v.integer, MyInt)
    assert v.integer == 1
    assert isinstance(v.floating, MyFloat)
    assert v.floating == 2.2
    assert isinstance(v.numeric, MyInt)
    assert v.numeric == 3


def test_annotable_picklable_with_additional_attributes():
    a = BetweenWithExtra(10, lower=2)
    b = BetweenWithExtra(10, lower=2)
    assert a == b
    assert a is not b

    a.extra = 1
    assert a.extra == 1
    assert a != b

    assert a == pickle.loads(pickle.dumps(a))


def test_annotable_is_mutable_by_default():
    # TODO(kszucs): more exhaustive testing of mutability, e.g. setting
    # optional value to None doesn't set to the default value
    class Op(Annotable):
        __slots__ = ("custom",)

        a = argument(is_int)
        b = argument(is_int)

    p = Op(1, 2)
    assert p.a == 1
    p.a = 3
    assert p.a == 3
    assert p == Op(a=3, b=2)

    # test that non-annotable attributes can be set as well
    p.custom = 1
    assert p.custom == 1


def test_annotable_with_type_annotations() -> None:
    class Op1(Annotable):
        foo: int
        bar: str = ""

    p = Op1(1)
    assert p.foo == 1
    assert p.bar == ""

    with pytest.raises(MatchError):

        class Op2(Annotable):
            bar: str = None

    class Op2(Annotable):
        bar: str | None = None

    op = Op2()
    assert op.bar is None


class RecursiveNode(Annotable):
    child: Optional[Self] = None


def test_annotable_with_self_typehint():
    node = RecursiveNode(RecursiveNode(RecursiveNode(None)))
    assert isinstance(node, RecursiveNode)
    assert isinstance(node.child, RecursiveNode)
    assert isinstance(node.child.child, RecursiveNode)
    assert node.child.child.child is None

    with pytest.raises(MatchError):
        RecursiveNode(1)


def test_annotable_with_recursive_generic_type_annotations():
    # testing cons list
    pattern = Pattern.from_typehint(List[Integer], allow_coercion=True)
    values = ["1", 2.0, 3]
    result = pattern.apply(values, {})
    expected = ConsList(1, ConsList(2, ConsList(3, EmptyList())))
    assert result == expected
    assert result[0] == 1
    assert result[1] == 2
    assert result[2] == 3
    assert len(result) == 3
    with pytest.raises(IndexError):
        result[3]

    # testing cons map
    pattern = Pattern.from_typehint(Map[Integer, Float], allow_coercion=True)
    values = {"1": 2, 3: "4.0", 5: 6.0}
    result = pattern.apply(values, {})
    expected = ConsMap((1, 2.0), ConsMap((3, 4.0), ConsMap((5, 6.0), EmptyMap())))
    assert result == expected
    assert result[1] == 2.0
    assert result[3] == 4.0
    assert result[5] == 6.0
    assert len(result) == 3
    with pytest.raises(KeyError):
        result[7]

    # testing both encapsulated in a class
    expr = MyExpr(a="1", b=["2.0", 3, True], c={"a": "1", "b": 2, "c": 3.0})
    assert expr.a == 1
    assert expr.b == ConsList(2.0, ConsList(3.0, ConsList(1.0, EmptyList())))
    assert expr.c == ConsMap(("a", 1), ConsMap(("b", 2), ConsMap(("c", 3), EmptyMap())))


def test_annotable_as_immutable():
    class AnnImm(Annotable, immutable=True):
        value = argument(is_int)
        lower = optional(is_int, default=0)
        upper = optional(is_int, default=None)

    assert AnnImm.__mro__ == (AnnImm, Immutable, Annotable, Abstract, object)

    obj = AnnImm(3, lower=0, upper=4)
    with pytest.raises(AttributeError):
        obj.value = 1


def test_annotable_equality_checks():
    class Between(Annotable):
        value = argument(is_int)
        lower = optional(is_int, default=0)
        upper = optional(is_int, default=None)

    a = Between(3, lower=0, upper=4)
    b = Between(3, lower=0, upper=4)
    c = Between(2, lower=0, upper=4)

    assert a == b
    assert b == a
    assert a != c
    assert c != a
    assert a.__eq__(b)
    assert not a.__eq__(c)


def test_maintain_definition_order():
    class Between(Annotable):
        value = argument(is_int)
        lower = optional(is_int, default=0)
        upper = optional(is_int, default=None)

    assert Between.__argnames__ == ("value", "lower", "upper")


def test_signature_inheritance():
    class IntBinop(Annotable):
        left = argument(is_int)
        right = argument(is_int)

    class FloatAddRhs(IntBinop):
        right = argument(is_float)

    class FloatAddClip(FloatAddRhs):
        left = argument(is_float)
        clip_lower = optional(is_int, default=0)
        clip_upper = optional(is_int, default=10)

    class IntAddClip(FloatAddClip, IntBinop):
        pass

    assert IntBinop.__signature__ == Signature(
        {
            "left": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_int),
            "right": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_int),
        }
    )

    assert FloatAddRhs.__signature__ == Signature(
        {
            "left": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_int),
            "right": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_float),
        }
    )

    assert FloatAddClip.__signature__ == Signature(
        {
            "left": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_float),
            "right": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_float),
            "clip_lower": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_int, default=0),
            "clip_upper": Parameter(
                Parameter.POSITIONAL_OR_KEYWORD, is_int, default=10
            ),
        }
    )

    assert IntAddClip.__signature__ == Signature(
        {
            "left": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_float),
            "right": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_float),
            "clip_lower": Parameter(Parameter.POSITIONAL_OR_KEYWORD, is_int, default=0),
            "clip_upper": Parameter(
                Parameter.POSITIONAL_OR_KEYWORD, is_int, default=10
            ),
        }
    )


def test_positional_argument_reordering():
    class Farm(Annotable):
        ducks = argument(is_int)
        donkeys = argument(is_int)
        horses = argument(is_int)
        goats = argument(is_int)
        chickens = argument(is_int)

    class NoHooves(Farm):
        horses = optional(is_int, default=0)
        goats = optional(is_int, default=0)
        donkeys = optional(is_int, default=0)

    f1 = Farm(1, 2, 3, 4, 5)
    f2 = Farm(1, 2, goats=4, chickens=5, horses=3)
    f3 = Farm(1, 0, 0, 0, 100)
    assert f1 == f2
    assert f1 != f3

    g1 = NoHooves(1, 2, donkeys=-1)
    assert g1.ducks == 1
    assert g1.chickens == 2
    assert g1.donkeys == -1
    assert g1.horses == 0
    assert g1.goats == 0


def test_keyword_argument_reordering():
    class Alpha(Annotable):
        a = argument(is_int)
        b = argument(is_int)

    class Beta(Alpha):
        c = argument(is_int)
        d = optional(is_int, default=0)
        e = argument(is_int)

    obj = Beta(1, 2, 3, 4)
    assert obj.a == 1
    assert obj.b == 2
    assert obj.c == 3
    assert obj.e == 4
    assert obj.d == 0

    obj = Beta(1, 2, 3, 4, 5)
    assert obj.d == 5
    assert obj.e == 4


def test_variadic_argument_reordering():
    class Test(Annotable):
        a = argument(is_int)
        b = argument(is_int)
        args = varargs(is_int)

    class Test2(Test):
        c = argument(is_int)
        args = varargs(is_int)

    with pytest.raises(TypeError, match="missing a required argument: 'c'"):
        Test2(1, 2)

    a = Test2(1, 2, 3)
    assert a.a == 1
    assert a.b == 2
    assert a.c == 3
    assert a.args == ()

    b = Test2(*range(5))
    assert b.a == 0
    assert b.b == 1
    assert b.c == 2
    assert b.args == (3, 4)

    msg = "only one variadic \\*args parameter is allowed"
    with pytest.raises(TypeError, match=msg):

        class Test3(Test):
            another_args = varargs(is_int)


def test_variadic_keyword_argument_reordering():
    class Test(Annotable):
        a = argument(is_int)
        b = argument(is_int)
        options = varkwargs(is_int)

    class Test2(Test):
        c = argument(is_int)
        options = varkwargs(is_int)

    with pytest.raises(TypeError, match="missing a required argument: 'c'"):
        Test2(1, 2)

    a = Test2(1, 2, c=3)
    assert a.a == 1
    assert a.b == 2
    assert a.c == 3
    assert a.options == {}

    b = Test2(1, 2, c=3, d=4, e=5)
    assert b.a == 1
    assert b.b == 2
    assert b.c == 3
    assert b.options == {"d": 4, "e": 5}

    msg = "only one variadic \\*\\*kwargs parameter is allowed"
    with pytest.raises(TypeError, match=msg):

        class Test3(Test):
            another_options = varkwargs(is_int)


def test_variadic_argument():
    class Test(Annotable):
        a = argument(is_int)
        b = argument(is_int)
        args = varargs(is_int)

    assert Test(1, 2).args == ()
    assert Test(1, 2, 3).args == (3,)
    assert Test(1, 2, 3, 4, 5).args == (3, 4, 5)


def test_variadic_keyword_argument():
    class Test(Annotable):
        first = argument(is_int)
        second = argument(is_int)
        options = varkwargs(is_int)

    assert Test(1, 2).options == {}
    assert Test(1, 2, a=3).options == {"a": 3}
    assert Test(1, 2, a=3, b=4, c=5).options == {"a": 3, "b": 4, "c": 5}


# def test_copy_with_variadic_argument():
#     class Foo(Annotable):
#         a = is_int
#         b = is_int
#         args = varargs(is_int)

#     class Bar(Concrete):
#         a = is_int
#         b = is_int
#         args = varargs(is_int)

#     for t in [Foo(1, 2, 3, 4, 5), Bar(1, 2, 3, 4, 5)]:
#         assert t.a == 1
#         assert t.b == 2
#         assert t.args == (3, 4, 5)

#         u = t.copy(a=6, args=(8, 9, 10))
#         assert u.a == 6
#         assert u.b == 2
#         assert u.args == (8, 9, 10)


# def test_concrete_copy_with_unknown_argument_raise():
#     class Bar(Concrete):
#         a = is_int
#         b = is_int

#     t = Bar(1, 2)
#     assert t.a == 1
#     assert t.b == 2

#     with pytest.raises(AttributeError, match="Unexpected arguments"):
#         t.copy(c=3, d=4)


def test_immutable_pickling_variadic_arguments():
    v = VariadicArgs(1, 2, 3, 4, 5)
    assert v.args == (1, 2, 3, 4, 5)
    assert v == pickle.loads(pickle.dumps(v))

    v = VariadicKeywords(a=3, b=4, c=5)
    assert v.kwargs == {"a": 3, "b": 4, "c": 5}
    assert v == pickle.loads(pickle.dumps(v))

    v = VariadicArgsAndKeywords(1, 2, 3, 4, 5, a=3, b=4, c=5)
    assert v.args == (1, 2, 3, 4, 5)
    assert v.kwargs == {"a": 3, "b": 4, "c": 5}
    assert v == pickle.loads(pickle.dumps(v))


def test_dont_copy_default_argument():
    default = tuple()

    class Op(Annotable):
        arg = optional(IsType(tuple), default=default)

    op = Op()
    assert op.arg is default


# def test_copy_mutable_with_default_attribute():
#     class Test(Annotable):
#         a = attribute(InstanceOf(dict), default={})
#         b = argument(InstanceOf(str))  # required argument

#         @attribute
#         def c(self):
#             return self.b.upper()

#     t = Test("t")
#     assert t.a == {}
#     assert t.b == "t"
#     assert t.c == "T"

#     with pytest.raises(ValidationError):
#         t.a = 1
#     t.a = {"map": "ping"}
#     assert t.a == {"map": "ping"}

#     assert t.copy() == t

#     u = t.copy(b="u")
#     assert u.b == "u"
#     assert u.c == "T"
#     assert u.a == {"map": "ping"}

#     x = t.copy(a={"emp": "ty"})
#     assert x.a == {"emp": "ty"}
#     assert x.b == "t"


def test_slots_are_inherited_and_overridable():
    class Op(Annotable):
        __slots__ = ("_cache",)  # first definition
        arg = argument(Anything())

    class StringOp(Op):
        arg = argument(IsType(str))  # new overridden slot

    class StringSplit(StringOp):
        sep = argument(IsType(str))  # new slot

    class StringJoin(StringOp):
        __slots__ = ("_memoize",)  # new slot
        sep = argument(IsType(str))  # new overridden slot

    assert Op.__slots__ == ("_cache", "arg")
    assert StringOp.__slots__ == ("arg",)
    assert StringSplit.__slots__ == ("sep",)
    assert StringJoin.__slots__ == ("_memoize", "sep")


def test_multiple_inheritance():
    # multiple inheritance is allowed only if one of the parents has non-empty
    # __slots__ definition, otherwise python will raise lay-out conflict

    class Op(Annotable):
        __slots__ = ("_hash",)

    class Value(Annotable):
        arg = argument(IsType(object))

    class Reduction(Value):
        pass

    class UDF(Value):
        func = argument(IsType(Callable))

    class UDAF(UDF, Reduction):
        arity = argument(is_int)

    class A(Annotable):
        a = argument(is_int)

    class B(Annotable):
        b = argument(is_int)

    msg = "multiple bases have instance lay-out conflict"
    with pytest.raises(TypeError, match=msg):

        class AB(A, B):
            ab = argument(is_int)

    assert UDAF.__slots__ == ("arity",)
    strlen = UDAF(arg=2, func=lambda value: len(str(value)), arity=1)
    assert strlen.arg == 2
    assert strlen.arity == 1


@pytest.mark.parametrize(
    "obj",
    [
        StringOp("something"),
        StringOp(arg="something"),
    ],
)
def test_pickling_support(obj):
    assert obj == pickle.loads(pickle.dumps(obj))


def test_multiple_inheritance_argument_order():
    class Value(Annotable):
        arg = argument(is_any)

    class VersionedOp(Value):
        version = argument(is_int)

    class Reduction(Annotable):
        pass

    class Sum(VersionedOp, Reduction):
        where = optional(is_bool, default=False)

    assert tuple(Sum.__signature__.parameters.keys()) == ("arg", "version", "where")


def test_multiple_inheritance_optional_argument_order():
    class Value(Annotable):
        pass

    class ConditionalOp(Annotable):
        where = optional(is_bool, default=False)

    class Between(Value, ConditionalOp):
        min = argument(is_int)
        max = argument(is_int)
        how = optional(is_str, default="strict")

    assert tuple(Between.__signature__.parameters.keys()) == (
        "min",
        "max",
        "how",
        "where",
    )


def test_immutability():
    class Value(Annotable, immutable=True):
        a = argument(is_int)

    op = Value(1)
    with pytest.raises(AttributeError):
        op.a = 3


class BaseValue(Annotable):
    i = argument(is_int)
    j = attribute(is_int)


class Value2(BaseValue):
    @attribute
    def k(self):
        return 3


class Value3(BaseValue):
    k = attribute(is_int, default=3)


class Value4(BaseValue):
    k = attribute(Option(is_int), default=None)


def test_annotable_with_dict_slot():
    class Flexible(Annotable):
        __slots__ = ("__dict__",)

    v = Flexible()
    v.a = 1
    v.b = 2
    assert v.a == 1
    assert v.b == 2


def test_annotable_attribute():
    with pytest.raises(TypeError, match="too many positional arguments"):
        BaseValue(1, 2)

    v = BaseValue(1)
    assert v.__slots__ == ("i", "j")
    assert v.i == 1
    assert not hasattr(v, "j")
    v.j = 2
    assert v.j == 2

    # TODO(kszucs)
    # with pytest.raises(TypeError):
    #     v.j = "foo"


def test_annotable_attribute_init():
    assert Value2.__slots__ == ("k",)
    v = Value2(1)

    assert v.i == 1
    assert not hasattr(v, "j")
    v.j = 2
    assert v.j == 2
    assert v.k == 3

    v = Value3(1)
    assert v.k == 3

    v = Value4(1)
    assert v.k is None


def test_annotable_mutability_and_serialization():
    v_ = BaseValue(1)
    v_.j = 2
    v = BaseValue(1)
    v.j = 2
    assert v_ == v
    assert v_.j == v.j == 2

    assert repr(v) == "BaseValue(i=1)"
    w = pickle.loads(pickle.dumps(v))
    assert w.i == 1
    assert w.j == 2
    assert v == w

    v.j = 4
    assert v_ != v
    w = pickle.loads(pickle.dumps(v))
    assert w == v
    assert repr(w) == "BaseValue(i=1)"


def test_initialized_attribute_basics():
    class Value(Annotable):
        a = argument(is_int)

        @attribute
        def double_a(self):
            return 2 * self.a

    op = Value(1)
    assert op.a == 1
    assert op.double_a == 2
    assert "double_a" in Value.__slots__


def test_initialized_attribute_with_validation():
    class Value(Annotable):
        a = argument(is_int)

        @attribute(int)
        def double_a(self):
            return 2 * self.a

    op = Value(1)
    assert op.a == 1
    assert op.double_a == 2
    assert "double_a" in Value.__slots__

    op.double_a = 3
    assert op.double_a == 3

    with pytest.raises(MatchError):
        op.double_a = "foo"


def test_initialized_attribute_mixed_with_classvar():
    class Value(Annotable):
        arg = argument(is_int)

        shape = "like-arg"
        dtype = "like-arg"

    class Reduction(Value):
        shape = "scalar"

    class Variadic(Value):
        @attribute
        def shape(self):
            if self.arg > 10:
                return "columnar"
            else:
                return "scalar"

    r = Reduction(1)
    assert r.shape == "scalar"
    assert "shape" not in r.__slots__

    v = Variadic(1)
    assert v.shape == "scalar"
    assert "shape" in v.__slots__

    v = Variadic(100)
    assert v.shape == "columnar"
    assert "shape" in v.__slots__


def test_hashable():
    assert BetweenWithCalculated.__mro__ == (
        BetweenWithCalculated,
        Hashable,
        Immutable,
        Annotable,
        Abstract,
        object,
    )

    assert BetweenWithCalculated.__eq__ is Hashable.__eq__
    assert BetweenWithCalculated.__argnames__ == ("value", "lower", "upper")

    # annotable
    obj = BetweenWithCalculated(10, lower=5, upper=15)
    obj2 = BetweenWithCalculated(10, lower=5, upper=15)
    assert obj.value == 10
    assert obj.lower == 5
    assert obj.upper == 15
    assert obj.calculated == 15
    assert obj == obj2
    assert obj is not obj2
    assert obj != (10, 5, 15)
    assert obj.__args__ == (10, 5, 15)
    # assert obj.args == (10, 5, 15)
    # assert obj.argnames == ("value", "lower", "upper")

    # immutable
    with pytest.raises(AttributeError):
        obj.value = 11

    # hashable
    assert {obj: 1}.get(obj) == 1

    # serializable
    assert pickle.loads(pickle.dumps(obj)) == obj


def test_init_subclass_keyword_arguments():
    class Test(Annotable):
        def __init_subclass__(cls, **kwargs):
            super().__init_subclass__()
            cls.kwargs = kwargs

    class Test2(Test, something="value", value="something"):
        pass

    assert Test2.kwargs == {"something": "value", "value": "something"}


def test_argument_order_using_optional_annotations():
    class Case1(Annotable):
        results: Optional[tuple[int, ...]] = ()
        default: Optional[int] = None

    class SimpleCase1(Case1):
        base: int
        cases: Optional[tuple[int, ...]] = ()

    class Case2(Annotable):
        results = optional(TupleOf(is_int), default=())
        default = optional(is_int)

    class SimpleCase2(Case2):
        base = argument(is_int)
        cases = optional(TupleOf(is_int), default=())

    assert (
        SimpleCase1.__argnames__
        == SimpleCase2.__argnames__
        == ("base", "cases", "results", "default")
    )


def test_annotable_with_optional_coercible_typehint():
    class Example(Annotable):
        value: Optional[MyInt] = None

    assert Example().value is None
    assert Example(None).value is None
    assert Example(1).value == 1
    assert isinstance(Example(1).value, MyInt)


# def test_error_message(snapshot):
#     class Example(Annotable):
#         a: int
#         b: int = 0
#         c: str = "foo"
#         d: Optional[float] = None
#         e: tuple[int, ...] = (1, 2, 3)
#         f: As[int] = 1

#     with pytest.raises(ValidationError) as exc_info:
#         Example("1", "2", "3", "4", "5", [])

#     # assert "Failed" in str(exc_info.value)

#     if sys.version_info >= (3, 11):
#         target = "error_message_py311.txt"
#     else:
#         target = "error_message.txt"
#     snapshot.assert_match(str(exc_info.value), target)


def test_abstract_meta():
    class Foo(Abstract):
        @abstractmethod
        def foo(self): ...

        @property
        @abstractmethod
        def bar(self): ...

    assert not issubclass(type(Foo), ABCMeta)
    assert issubclass(type(Foo), AbstractMeta)
    assert Foo.__abstractmethods__ == frozenset({"foo", "bar"})

    with pytest.raises(TypeError, match="Can't instantiate abstract class .*Foo.*"):
        Foo()

    class Bar(Foo):
        def foo(self):
            return 1

        @property
        def bar(self):
            return 2

    bar = Bar()
    assert bar.foo() == 1
    assert bar.bar == 2
    assert isinstance(bar, Foo)
    assert Bar.__abstractmethods__ == frozenset()


def test_annotable_supports_abstractmethods():
    class Foo(Annotable):
        @abstractmethod
        def foo(self): ...

        @property
        @abstractmethod
        def bar(self): ...

    assert not issubclass(type(Foo), ABCMeta)
    assert issubclass(type(Foo), AnnotableMeta)
    assert Foo.__abstractmethods__ == frozenset({"foo", "bar"})

    with pytest.raises(TypeError, match="Can't instantiate abstract class .*Foo.*"):
        Foo()

    class Bar(Foo):
        def foo(self):
            return 1

        @property
        def bar(self):
            return 2

    bar = Bar()
    assert bar.foo() == 1
    assert bar.bar == 2
    assert isinstance(bar, Foo)
    assert isinstance(bar, Annotable)
    assert Bar.__abstractmethods__ == frozenset()


def test_annotable_recalculates_inherited_abstractmethods():
    class Abstract(Annotable):
        @abstractmethod
        def foo(self): ...

        @property
        @abstractmethod
        def bar(self): ...

    class Mixin:
        def foo(self):
            return 1

        def bar(self):
            return 2

    class Foo(Mixin, Abstract):
        pass

    assert Abstract.__abstractmethods__ == frozenset({"foo", "bar"})
    assert Foo.__abstractmethods__ == frozenset()


# TODO(kszucs): test __new__ as well
def test_annotable_with_custom_init():
    called_with = None

    class MyInit(Annotable):
        a = argument(int)
        b = argument(IsType(str))
        c = optional(float, default=0.0)

        def __init__(self, a, b, c):
            nonlocal called_with
            called_with = (a, b, c)
            super().__init__(a=a, b=b, c=c)

        @attribute
        def called_with(self):
            return (self.a, self.b, self.c)

    assert MyInit.__spec__.initable is True
    with pytest.raises(MatchError):
        MyInit(1, 2, 3)

    mi = MyInit(1, "2", 3.3)
    assert called_with == (1, "2", 3.3)
    assert isinstance(mi, MyInit)
    assert mi.a == 1
    assert mi.b == "2"
    assert mi.c == 3.3
    assert mi.called_with == called_with


def test_any():
    class MyClass(Annotable):
        foo: Any

    assert MyClass(1).foo == 1


class Bar(Annotable):
    x: int


def test_nested():
    class Foo(Annotable):
        bar: Bar

    assert Foo(Bar(1)).bar.x == 1

    class Quux(Annotable):
        bar: Bar = Bar(2)

    assert Quux().bar.x == 2
    assert Quux(Bar(3)).bar.x == 3


def test_annotable_spec_flags():
    assert Annotable.__spec__.initable is False
    assert Annotable.__spec__.immutable is False
    assert Annotable.__spec__.hashable is False


def test_annotable_spec_flag_inheritance():
    class A(Annotable):
        pass

    class B(Annotable, immutable=True):
        pass

    assert A.__spec__.initable is False
    assert A.__spec__.immutable is False
    assert A.__spec__.hashable is False
    assert B.__spec__.initable is False
    assert B.__spec__.immutable is True
    assert B.__spec__.hashable is False

    class C(A, B):
        pass

    assert C.__spec__.initable is False
    assert C.__spec__.immutable is True
    assert C.__spec__.hashable is False

    class D(B, A, hashable=True):
        pass

    assert D.__spec__.initable is False
    assert D.__spec__.immutable is True
    assert D.__spec__.hashable is True

    with pytest.raises(TypeError):

        class E(A, B, immutable=False):
            pass

    with pytest.raises(TypeError):

        class F(D, hashable=False):
            pass

    with pytest.raises(TypeError):

        class G(D, immutable=False):
            pass


def test_user_model():
    class User(Annotable):
        id: int
        name: str = "Jane Doe"
        age: int | None = None
        children: list[str] = []

    assert User.__spec__.initable is False
    assert User.__spec__.immutable is False
    assert User.__spec__.hashable is False


def test_arg_and_hash_precomputed_before_attributes():
    class Frozen(Annotable, immutable=True, hashable=True):
        arg: int

        @attribute
        def a(self):
            assert self.__args__ == (1,)
            assert isinstance(self.__precomputed_hash__, int)
            return "ok"

    assert Frozen(1).a == "ok"


class OneAndOnly(Annotable, singleton=True):
    __instances__ = weakref.WeakValueDictionary()


class DataType(Annotable, singleton=True):
    __instances__ = weakref.WeakValueDictionary()
    nullable: bool = True


def test_singleton_basics():
    one = OneAndOnly()
    only = OneAndOnly()
    assert one is only

    assert len(OneAndOnly.__instances__) == 1
    key = (OneAndOnly,)
    assert OneAndOnly.__instances__[key] is one


def test_singleton_lifetime() -> None:
    one = OneAndOnly()
    assert len(OneAndOnly.__instances__) == 1

    del one
    assert len(OneAndOnly.__instances__) == 0


def test_singleton_with_argument() -> None:
    dt1 = DataType(nullable=True)
    dt2 = DataType(nullable=False)
    dt3 = DataType(nullable=True)

    assert dt1 is dt3
    assert dt1 is not dt2
    assert len(DataType.__instances__) == 2

    del dt3
    assert len(DataType.__instances__) == 2
    del dt1
    assert len(DataType.__instances__) == 1
    del dt2
    assert len(DataType.__instances__) == 0


def test_singleton_looked_after_validation() -> None:
    class Single(Annotable, singleton=True):
        value: As[int]

    # arguments looked up after validation
    obj1 = Single("1")
    obj2 = Single(2)
    assert Single("1") is obj1
    assert Single(1) is obj1
    assert Single(1.0) is obj1
    assert Single(2) is obj2
    assert Single("2") is obj2
    assert obj2 is not obj1
    assert Single("3") is Single(3.0)



from __future__ import annotations

import functools
import inspect
import typing
from collections.abc import Mapping, Sequence
from types import FunctionType, MethodType
from typing import Any, ClassVar, Optional
from weakref import WeakValueDictionary

import cython

from .patterns import (
    FrozenDictOf,
    Option,
    Pattern,
    PatternMap,
    TupleOf,
    _any,
    pattern,
)
from .utils import PseudoHashable, get_type_hints, get_type_origin

EMPTY = inspect.Parameter.empty
_ensure_pattern = pattern


@cython.final
@cython.cclass
class Attribute:
    pattern = cython.declare(Pattern, visibility="readonly")
    default_ = cython.declare(object, visibility="readonly")

    def __init__(self, pattern: Any = _any, default: Any = EMPTY):
        self.pattern = _ensure_pattern(pattern)
        self.default_ = default

    def __repr__(self):
        return f"<{self.__class__.__name__} pattern={self.pattern!r} default={self.default_!r}>"

    def __hash__(self) -> int:
        return hash((self.__class__, self.pattern, PseudoHashable(self.default_)))

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Attribute):
            return NotImplemented
        right: Attribute = cython.cast(Attribute, other)
        return self.pattern == right.pattern and self.default_ == right.default_

    def __call__(self, default):
        """Needed to support the decorator syntax."""
        return self.__class__(self.pattern, default)


_POSITIONAL_ONLY = cython.declare(cython.int, int(inspect.Parameter.POSITIONAL_ONLY))
_POSITIONAL_OR_KEYWORD = cython.declare(
    cython.int, int(inspect.Parameter.POSITIONAL_OR_KEYWORD)
)
_VAR_POSITIONAL = cython.declare(cython.int, int(inspect.Parameter.VAR_POSITIONAL))
_KEYWORD_ONLY = cython.declare(cython.int, int(inspect.Parameter.KEYWORD_ONLY))
_VAR_KEYWORD = cython.declare(cython.int, int(inspect.Parameter.VAR_KEYWORD))


@cython.final
@cython.cclass
class Parameter:
    POSITIONAL_ONLY: typing.ClassVar[int] = _POSITIONAL_ONLY
    POSITIONAL_OR_KEYWORD: typing.ClassVar[int] = _POSITIONAL_OR_KEYWORD
    VAR_POSITIONAL: typing.ClassVar[int] = _VAR_POSITIONAL
    KEYWORD_ONLY: typing.ClassVar[int] = _KEYWORD_ONLY
    VAR_KEYWORD: typing.ClassVar[int] = _VAR_KEYWORD

    kind = cython.declare(cython.int, visibility="readonly")
    pattern = cython.declare(Pattern, visibility="readonly")
    default_ = cython.declare(object, visibility="readonly")
    typehint = cython.declare(object, visibility="readonly")

    def __init__(
        self,
        kind: int,
        pattern: Any = _any,
        default: Any = EMPTY,
        typehint: Any = EMPTY,
    ):
        self.kind = kind
        self.typehint = typehint
        if kind is _VAR_POSITIONAL:
            self.pattern = TupleOf(pattern)
        elif kind is _VAR_KEYWORD:
            # TODO(kszucs): remove FrozenDict?
            self.pattern = FrozenDictOf(_any, pattern)
        else:
            self.pattern = _ensure_pattern(pattern)

        # validate that the default value matches the pattern
        if default is not EMPTY:
            # TODO(kszucs): try/except MatchError raise an error indicating that the default value doesn't match the pattern
            self.default_ = self.pattern.match(default, {})
        else:
            self.default_ = default

    def format(self, name) -> str:
        result: str = name
        if self.typehint is not EMPTY:
            if hasattr(self.typehint, "__qualname__"):
                result += f": {self.typehint.__qualname__}"
            else:
                result += f": {self.typehint}"
        if self.default_ is not EMPTY:
            if self.typehint is EMPTY:
                result = f"{result}={self.default_}"
            else:
                result = f"{result} = {self.default_!r}"
        if self.kind == _VAR_POSITIONAL:
            result = f"*{result}"
        elif self.kind == _VAR_KEYWORD:
            result = f"**{result}"
        return result

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Parameter):
            return NotImplemented
        right: Parameter = cython.cast(Parameter, other)
        return (
            self.kind == right.kind
            and self.default_ == right.default_
            and self.typehint == right.typehint
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.__class__,
                self.kind,
                self.pattern,
                PseudoHashable(self.default_),
                self.typehint,
            )
        )


@cython.final
@cython.cclass
class Signature:
    length = cython.declare(cython.int, visibility="readonly")
    parameters = cython.declare(dict[str, Parameter], visibility="readonly")
    return_pattern = cython.declare(Pattern, visibility="readonly")
    return_typehint = cython.declare(object, visibility="readonly")

    def __init__(
        self,
        parameters: dict[str, Parameter],
        return_pattern: Pattern = _any,
        return_typehint: Any = EMPTY,
    ):
        self.length = len(parameters)
        self.parameters = parameters
        self.return_pattern = return_pattern
        self.return_typehint = return_typehint

    @staticmethod
    def from_callable(
        func: Any,
        arg_patterns: Sequence[Any] | Mapping[str, Any] | None = None,
        return_pattern: Any = None,
        allow_coercion: bool = True,
    ) -> Signature:
        sig = inspect.signature(func)
        hints = get_type_hints(func)
        params: dict[str, Parameter] = {}

        if arg_patterns is None:
            arg_patterns = {}
        elif isinstance(arg_patterns, Sequence):
            # create a mapping of parameter name to pattern
            arg_patterns = dict(zip(sig.parameters.keys(), arg_patterns))
        elif not isinstance(arg_patterns, Mapping):
            raise TypeError("arg_patterns must be a sequence or a mapping")

        for name, param in sig.parameters.items():
            typehint = hints.get(name, EMPTY)
            if name in arg_patterns:
                argpat = pattern(arg_patterns[name])
            elif typehint is not EMPTY:
                argpat = Pattern.from_typehint(typehint, allow_coercion=allow_coercion)
            else:
                argpat = _any

            params[name] = Parameter(
                kind=int(param.kind),
                default=param.default,
                pattern=argpat,
                typehint=typehint,
            )

        return_typehint = hints.get("return", EMPTY)
        if return_pattern is not None:
            retpat = pattern(return_pattern)
        elif return_typehint is not EMPTY:
            retpat = Pattern.from_typehint(
                return_typehint, allow_coercion=allow_coercion
            )
        else:
            retpat = _any

        return Signature(params, return_typehint=return_typehint, return_pattern=retpat)

    @staticmethod
    def merge(
        signatures: Sequence[Signature],
        parameters: Optional[dict[str, Parameter]] = None,
    ):
        """Merge multiple signatures.

        In addition to concatenating the parameters, it also reorders the
        parameters so that optional arguments come after mandatory arguments.

        Parameters
        ----------
        signatures :
            Signature instances to merge.
        parameters :
            Parameters to add to the merged signature.

        Returns
        -------
        Signature

        """
        name: str
        param: Parameter
        params: dict[str, Parameter] = {}
        for sig in signatures:
            params.update(sig.parameters)

        inherited: set[str] = set(params.keys())
        if parameters:
            for name, param in parameters.items():
                params[name] = param

        # mandatory fields without default values must precede the optional
        # ones in the function signature, the partial ordering will be kept
        old_args: list = []
        new_args: list = []
        var_args: list = []
        var_kwargs: list = []
        new_kwargs: list = []
        old_kwargs: list = []
        for name, param in params.items():
            if param.kind == _VAR_POSITIONAL:
                if var_args:
                    raise TypeError("only one variadic *args parameter is allowed")
                var_args.append(name)
            elif param.kind == _VAR_KEYWORD:
                if var_kwargs:
                    raise TypeError("only one variadic **kwargs parameter is allowed")
                var_kwargs.append(name)
            elif name in inherited:
                if param.default_ is EMPTY:
                    old_args.append(name)
                else:
                    old_kwargs.append(name)
            elif param.default_ is EMPTY:
                new_args.append(name)
            else:
                new_kwargs.append(name)

        order: list[str] = (
            old_args + new_args + var_args + new_kwargs + old_kwargs + var_kwargs
        )
        return Signature({name: params[name] for name in order})

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Signature):
            return NotImplemented
        right: Signature = cython.cast(Signature, other)
        return (
            tuple(self.parameters.items()) == tuple(right.parameters.items())
            and self.return_pattern == right.return_pattern
            and self.return_typehint == right.return_typehint
        )

    def __hash__(self) -> int:
        return hash(
            (
                self.__class__,
                PseudoHashable(self.parameters),
                self.return_pattern,
                self.return_typehint,
            )
        )

    def __call__(self, /, *args, **kwargs):
        return self.bind(args, kwargs)

    def __len__(self) -> int:
        return self.length

    def __str__(self):
        params_str = ", ".join(
            param.format(name) for name, param in self.parameters.items()
        )
        if self.return_typehint is not EMPTY:
            return_str = f" -> {self.return_typehint}"
        else:
            return_str = ""
        return f"({params_str}){return_str}"

    @cython.ccall
    def bind(self, args: tuple[Any, ...], kwargs: dict[str, Any]) -> dict[str, Any]:
        """Bind the arguments to the signature.

        Parameters
        ----------
        args : Any
            Positional arguments.
        kwargs : Any
            Keyword arguments.

        Returns
        -------
        dict
            Mapping of parameter names to argument values.

        """
        i: cython.int = 0
        kind: cython.int
        param: Parameter
        bound: dict[str, Any] = {}
        params = iter(self.parameters.items())

        # 1. HANDLE ARGS
        for i in range(len(args)):
            try:
                name, param = next(params)
            except StopIteration:
                raise TypeError("too many positional arguments")

            kind = param.kind
            if kind is _POSITIONAL_OR_KEYWORD:
                if name in kwargs:
                    raise TypeError(f"multiple values for argument '{name}'")
                bound[name] = args[i]
            elif kind is _VAR_KEYWORD or kind is _KEYWORD_ONLY:
                raise TypeError("too many positional arguments")
            elif kind is _VAR_POSITIONAL:
                bound[name] = args[i:]
                break
            elif kind is _POSITIONAL_ONLY:
                bound[name] = args[i]
            else:
                raise TypeError("unreachable code")

        # 2. HANDLE KWARGS
        while True:
            try:
                name, param = next(params)
            except StopIteration:
                if kwargs:
                    raise TypeError(
                        f"got an unexpected keyword argument '{next(iter(kwargs))}'"
                    )
                break

            kind = param.kind
            if kind is _POSITIONAL_OR_KEYWORD or kind is _KEYWORD_ONLY:
                if name in kwargs:
                    bound[name] = kwargs.pop(name)
                elif param.default_ is EMPTY:
                    raise TypeError(f"missing a required argument: '{name}'")
                else:
                    bound[name] = param.default_
            elif kind is _VAR_POSITIONAL:
                bound[name] = ()
            elif kind is _VAR_KEYWORD:
                bound[name] = kwargs
                break
            elif kind is _POSITIONAL_ONLY:
                if param.default_ is EMPTY:
                    if name in kwargs:
                        raise TypeError(
                            f"positional only argument '{name}' passed as keyword argument"
                        )
                    else:
                        raise TypeError(f"missing required positional argument {name}")
                else:
                    bound[name] = param.default_
            else:
                raise TypeError("unreachable code")

        return bound

    @cython.ccall
    def unbind(self, bound: dict[str, Any]) -> tuple[tuple[Any, ...], dict[str, Any]]:
        """Reverse bind of the parameters.

        Attempts to reconstructs the original arguments as keyword only arguments.

        Parameters
        ----------
        bound
            Object with attributes matching the signature parameters.

        Returns
        -------
        args : (args, kwargs)
            Tuple of positional and keyword arguments.

        """
        # does the reverse of bind, but doesn't apply defaults
        args: list = []
        kind: cython.int
        kwargs: dict = {}
        param: Parameter
        for name, param in self.parameters.items():
            value = bound[name]
            kind = param.kind
            if kind is _POSITIONAL_OR_KEYWORD:
                args.append(value)
            elif kind is _VAR_POSITIONAL:
                args.extend(value)
            elif kind is _VAR_KEYWORD:
                kwargs.update(value)
            elif kind is _KEYWORD_ONLY:
                kwargs[name] = value
            elif kind is _POSITIONAL_ONLY:
                args.append(value)
            else:
                raise TypeError(f"unsupported parameter kind {kind}")

        return tuple(args), kwargs


def annotated(_1=None, _2=None, _3=None, **kwargs):
    """Create functions with arguments validated at runtime.

    There are various ways to apply this decorator:

    1. With type annotations

    >>> @annotated
    ... def foo(x: int, y: str) -> float:
    ...     return float(x) + float(y)

    2. With argument patterns passed as keyword arguments

    >>> from ibis.common.patterns import InstanceOf as instance_of
    >>> @annotated(x=instance_of(int), y=instance_of(str))
    ... def foo(x, y):
    ...     return float(x) + float(y)

    3. With mixing type annotations and patterns where the latter takes precedence

    >>> @annotated(x=instance_of(float))
    ... def foo(x: int, y: str) -> float:
    ...     return float(x) + float(y)

    4. With argument patterns passed as a list and/or an optional return pattern

    >>> @annotated([instance_of(int), instance_of(str)], instance_of(float))
    ... def foo(x, y):
    ...     return float(x) + float(y)

    Parameters
    ----------
    *args : Union[
                tuple[Callable],
                tuple[list[Pattern], Callable],
                tuple[list[Pattern], Pattern, Callable]
            ]
        Positional arguments.
        - If a single callable is passed, it's wrapped with the signature
        - If two arguments are passed, the first one is a list of patterns for the
          arguments and the second one is the callable to wrap
        - If three arguments are passed, the first one is a list of patterns for the
          arguments, the second one is a pattern for the return value and the third
          one is the callable to wrap
    **kwargs : dict[str, Pattern]
        Patterns for the arguments.

    Returns
    -------
    Callable

    """
    if _1 is None:
        return functools.partial(annotated, **kwargs)
    elif _2 is None:
        if callable(_1):
            func, patterns, return_pattern = _1, None, None
        else:
            return functools.partial(annotated, _1, **kwargs)
    elif _3 is None:
        if not isinstance(_2, Pattern):
            func, patterns, return_pattern = _2, _1, None
        else:
            return functools.partial(annotated, _1, _2, **kwargs)
    else:
        func, patterns, return_pattern = _3, _1, _2

    sig: Signature = Signature.from_callable(
        func,
        arg_patterns=patterns or kwargs,
        return_pattern=return_pattern,
        allow_coercion=True,
    )
    pat: Pattern = PatternMap(
        {name: param.pattern for name, param in sig.parameters.items()}
    )

    @functools.wraps(func)
    def wrapped(*args, **kwargs):
        # 0. Bind the arguments to the signature
        bound: dict[str, Any] = sig.bind(args, kwargs)

        # 1. Validate the passed arguments
        values: Any = pat.match(bound, {})

        # 2. Reconstruction of the original arguments
        args, kwargs = sig.unbind(values)

        # 3. Call the function with the validated arguments
        result = func(*args, **kwargs)

        # 4. Validate the return value
        result = sig.return_pattern.match(result, {})

        return result

    wrapped.__signature__ = sig

    return wrapped


def attribute(pattern=_any, default=EMPTY):
    """Annotation to mark a field in a class."""
    if default is EMPTY and isinstance(pattern, (FunctionType, MethodType)):
        return Attribute(pattern=_any, default=pattern)
    else:
        return Attribute(pattern=pattern, default=default)


def argument(pattern=_any, default=EMPTY, typehint=EMPTY):
    """Annotation type for all fields which should be passed as arguments."""
    return Parameter(
        kind=_POSITIONAL_OR_KEYWORD, default=default, pattern=pattern, typehint=typehint
    )


def optional(pattern=_any, default=None, typehint=EMPTY):
    """Annotation to allow and treat `None` values as missing arguments."""
    if default is None:
        pattern = Option(pattern)
    return Parameter(
        kind=_POSITIONAL_OR_KEYWORD, default=default, pattern=pattern, typehint=typehint
    )


def varargs(pattern=_any, typehint=EMPTY):
    """Annotation to mark a variable length positional arguments."""
    return Parameter(kind=_VAR_POSITIONAL, pattern=pattern, typehint=typehint)


def varkwargs(pattern=_any, typehint=EMPTY):
    """Annotation to mark a variable length keyword arguments."""
    return Parameter(kind=_VAR_KEYWORD, pattern=pattern, typehint=typehint)


__type_call__ = cython.declare(object, type.__call__)
if cython.compiled:
    from cython.cimports.cpython.object import PyObject_GenericSetAttr as __setattr__
else:
    __setattr__ = object.__setattr__


@cython.final
@cython.cclass
class AnnotableSpec:
    # make them readonly
    initable = cython.declare(cython.bint, visibility="readonly")
    hashable = cython.declare(cython.bint, visibility="readonly")
    immutable = cython.declare(cython.bint, visibility="readonly")
    singleton = cython.declare(cython.bint, visibility="readonly")
    signature = cython.declare(Signature, visibility="readonly")
    attributes = cython.declare(dict[str, Attribute], visibility="readonly")
    hasattribs = cython.declare(cython.bint, visibility="readonly")

    def __init__(
        self,
        initable: bool,
        hashable: bool,
        immutable: bool,
        singleton: bool,
        signature: Signature,
        attributes: dict[str, Attribute],
    ):
        self.initable = initable
        self.hashable = hashable
        self.immutable = immutable
        self.singleton = singleton
        self.signature = signature
        self.attributes = attributes
        self.hasattribs = bool(attributes)

    @cython.cfunc
    @cython.inline
    def new(self, cls: type, args: tuple[Any, ...], kwargs: dict[str, Any]):
        bound: dict[str, Any]
        if not args and len(kwargs) == self.signature.length:
            bound = kwargs
        else:
            bound = self.signature.bind(args, kwargs)

        if self.singleton or self.initable:
            return self.new_slow(cls, bound)
        else:
            return self.new_fast(cls, bound)

    @cython.cfunc
    @cython.inline
    def new_slow(self, cls: type, bound: dict[str, Any]):
        # slow initialization calling __init__
        ctx: dict[str, Any] = {}
        param: Parameter
        for name, param in self.signature.parameters.items():
            bound[name] = param.pattern.match(bound[name], ctx)

        if self.singleton:
            key = (cls, *bound.items())
            try:
                return cls.__instances__[key]
            except KeyError:
                this = __type_call__(cls, **bound)
                cls.__instances__[key] = this
                return this

        return __type_call__(cls, **bound)

    @cython.cfunc
    @cython.inline
    def new_fast(self, cls: type, bound: dict[str, Any]):
        # fast initialization directly setting the arguments
        ctx: dict[str, Any] = {}
        param: Parameter
        this = cls.__new__(cls)
        for name, param in self.signature.parameters.items():
            __setattr__(this, name, param.pattern.match(bound[name], ctx))
        if self.hashable:
            self.init_precomputes(this)
        if self.hasattribs:
            self.init_attributes(this)
        return this

    @cython.cfunc
    @cython.inline
    def init_attributes(self, this) -> cython.void:
        attr: Attribute
        for name, attr in self.attributes.items():
            if attr.default_ is not EMPTY:
                if callable(attr.default_):
                    value = attr.default_(this)
                else:
                    value = attr.default_
                __setattr__(this, name, value)

    @cython.cfunc
    @cython.inline
    def init_precomputes(self, this) -> cython.void:
        arguments = tuple(getattr(this, name) for name in self.signature.parameters)
        hashvalue = hash((this.__class__, arguments))
        __setattr__(this, "__args__", arguments)
        __setattr__(this, "__precomputed_hash__", hashvalue)


class AbstractMeta(type):
    """Base metaclass for many of the ibis core classes.

    Enforce the subclasses to define a `__slots__` attribute.

    Support abstract methods without extending `abc.ABCMeta`. While it provides
    a reduced feature set compared to `abc.ABCMeta` (no way to register virtual
    subclasses) but avoids expensive instance checks by enforcing explicit
    subclassing.
    """

    __slots__ = ()

    def __new__(metacls, clsname, bases, dct, **kwargs):
        # enforce slot definitions
        dct.setdefault("__slots__", ())

        # construct the class object
        cls = super().__new__(metacls, clsname, bases, dct, **kwargs)

        # calculate abstract methods existing in the class
        abstracts = {
            name
            for name, value in dct.items()
            if getattr(value, "__isabstractmethod__", False)
        }
        for parent in bases:
            for name in getattr(parent, "__abstractmethods__", set()):
                value = getattr(cls, name, None)
                if getattr(value, "__isabstractmethod__", False):
                    abstracts.add(name)

        # set the abstract methods for the class
        cls.__abstractmethods__ = frozenset(abstracts)

        return cls


class Abstract(metaclass=AbstractMeta):
    """Base class for many of the ibis core classes, see `AbstractMeta`."""


class AnnotableMeta(AbstractMeta):
    def __new__(
        metacls,
        clsname,
        bases,
        dct,
        initable=None,
        hashable=None,
        immutable=None,
        singleton=False,
        allow_coercion=True,
        **kwargs,
    ):
        # inherit annotable specifications from parent classes
        spec: AnnotableSpec
        signatures: list = []
        attributes: dict[str, Attribute] = {}
        is_initable: cython.bint
        is_hashable: cython.bint = hashable is True
        is_immutable: cython.bint = immutable is True
        is_singleton: cython.bint = singleton is True
        if initable is None:
            is_initable = "__init__" in dct or "__new__" in dct
        else:
            is_initable = initable
        for parent in bases:
            try:  # noqa: SIM105
                spec = parent.__spec__
            except AttributeError:
                continue
            is_initable |= spec.initable
            is_hashable |= spec.hashable
            is_immutable |= spec.immutable
            signatures.append(spec.signature)
            attributes.update(spec.attributes)

        # create the base classes for the new class
        traits: list[type] = []
        if is_immutable and immutable is False:
            raise TypeError(
                "One of the base classes is immutable so the child class cannot be mutable"
            )
        if is_hashable and hashable is False:
            raise TypeError(
                "One of the base classes is hashable so this child class must be hashable"
            )
        if is_hashable and not is_immutable:
            raise TypeError("Only immutable classes can be hashable")
        if hashable:
            traits.append(Hashable)
        if immutable:
            traits.append(Immutable)
        if singleton:
            traits.append(Singleton)

        # collect type annotations and convert them to patterns
        slots: list[str] = list(dct.pop("__slots__", []))
        module: str | None = dct.pop("__module__", None)
        qualname: str = dct.pop("__qualname__", clsname)
        annotations: dict[str, Any] = dct.get("__annotations__", {})
        if module is None:
            self_qualname = None
        else:
            self_qualname = f"{module}.{qualname}"

        # TODO(kszucs): pass dct as localns to get_type_hints
        typehints = get_type_hints(annotations, module=module)
        for name, typehint in typehints.items():
            if get_type_origin(typehint) is ClassVar:
                continue
            dct[name] = Parameter(
                kind=_POSITIONAL_OR_KEYWORD,
                pattern=Pattern.from_typehint(
                    typehint, allow_coercion=allow_coercion, self_qualname=self_qualname
                ),
                default=dct.get(name, EMPTY),
                typehint=typehint,
            )

        namespace: dict[str, Any] = {}
        parameters: dict[str, Parameter] = {}
        for name, value in dct.items():
            if isinstance(value, Parameter):
                parameters[name] = value
                slots.append(name)
            elif isinstance(value, Attribute):
                attributes[name] = value
                slots.append(name)
            else:
                namespace[name] = value

        # merge the annotations with the parent annotations
        signature = Signature.merge(signatures, parameters)
        argnames = tuple(signature.parameters.keys())
        bases = tuple(traits) + bases
        spec = AnnotableSpec(
            initable=is_initable,
            hashable=is_hashable,
            singleton=is_singleton,
            immutable=is_immutable,
            signature=signature,
            attributes=attributes,
        )

        namespace.update(
            __argnames__=argnames,
            __match_args__=argnames,
            __module__=module,
            __qualname__=qualname,
            __signature__=signature,
            __slots__=tuple(slots),
            __spec__=spec,
        )
        return super().__new__(metacls, clsname, bases, namespace, **kwargs)

    def __call__(cls, *args, **kwargs):
        spec: AnnotableSpec = cython.cast(AnnotableSpec, cls.__spec__)
        return spec.new(cython.cast(type, cls), args, kwargs)


class Singleton(Abstract):
    """Cache instances of the class based on instantiation arguments."""

    __instances__: Mapping[Any, Self] = WeakValueDictionary()
    __slots__ = ("__weakref__",)


class Immutable(Abstract):
    def __copy__(self):
        return self

    def __deepcopy__(self, memo):
        return self

    def __setattr__(self, name: str, _: Any) -> None:
        raise AttributeError(
            f"Attribute {name!r} cannot be assigned to immutable instance of "
            f"type {type(self)}"
        )


class Hashable(Abstract):
    __slots__ = ("__args__", "__precomputed_hash__")

    def __hash__(self) -> int:
        return self.__precomputed_hash__

    def __eq__(self, other) -> bool:
        if type(self) is not type(other):
            return NotImplemented
        return (
            self.__precomputed_hash__ == other.__precomputed_hash__
            and self.__args__ == other.__args__
        )


class Annotable(Abstract, metaclass=AnnotableMeta, initable=False):
    __argnames__: ClassVar[tuple[str, ...]]
    __match_args__: ClassVar[tuple[str, ...]]
    __signature__: ClassVar[Signature]

    def __init__(self, **kwargs):
        spec: AnnotableSpec = self.__spec__
        for name, value in kwargs.items():
            __setattr__(self, name, value)
        if spec.hashable:
            spec.init_precomputes(self)
        if spec.hasattribs:
            spec.init_attributes(self)

    def __setattr__(self, name, value) -> None:
        spec: AnnotableSpec = self.__spec__
        attr: Attribute
        param: Parameter
        if param := spec.signature.parameters.get(name):
            # try to look up the parameter
            value = param.pattern.match(value, {})
        elif attr := spec.attributes.get(name):
            # try to look up the attribute
            value = attr.pattern.match(value, {})
        __setattr__(self, name, value)

    def __eq__(self, other):
        spec: AnnotableSpec = self.__spec__
        # compare types
        if type(self) is not type(other):
            return NotImplemented
        # compare parameters
        for name in spec.signature.parameters:
            if getattr(self, name) != getattr(other, name):
                return False
        # compare attributes
        for name in spec.attributes:
            if getattr(self, name, EMPTY) != getattr(other, name, EMPTY):
                return False
        return True

    def __getstate__(self):
        spec: AnnotableSpec = self.__spec__
        state: dict[str, Any] = {}
        for name in spec.signature.parameters:
            state[name] = getattr(self, name)
        for name in spec.attributes:
            value = getattr(self, name, EMPTY)
            if value is not EMPTY:
                state[name] = value
        return state

    def __setstate__(self, state):
        spec: AnnotableSpec = self.__spec__
        for name, value in state.items():
            __setattr__(self, name, value)
        if spec.hashable:
            spec.init_precomputes(self)

    def __repr__(self) -> str:
        args = (f"{n}={getattr(self, n)!r}" for n in self.__argnames__)
        argstring = ", ".join(args)
        return f"{self.__class__.__name__}({argstring})"

    @property
    def __args__(self) -> tuple[Any, ...]:
        return tuple(getattr(self, name) for name in self.__argnames__)



from __future__ import annotations

import collections.abc
import functools
import inspect
import operator
from typing import Any, Optional

import cython

from .utils import PseudoHashable

Context = dict[str, Any]


@cython.cclass
class Deferred:
    """The user facing wrapper object providing syntactic sugar for deferreds.

    Provides a natural-like syntax for constructing deferred expressions by
    overloading all of the available dunder methods including the equality
    operator.

    Its sole purpose is to provide a nicer syntax for constructing deferred
    expressions, thus it gets unwrapped to the underlying deferred expression
    when used by the rest of the library.

    Parameters
    ----------
    deferred
        The deferred object to provide syntax sugar for.
    repr
        An optional fixed string to use when repr-ing the deferred expression,
        instead of the default. This is useful for complex deferred expressions
        where the arguments don't necessarily make sense to be user facing in
        the repr.
    """

    _repr: str
    _builder: Builder

    def __init__(self, builder: Builder, repr: Optional[str] = None):
        self._repr = repr
        self._builder = builder

    def __repr__(self):
        if self._repr is None:
            return repr(self._builder)
        else:
            return self._repr

    def __getattr__(self, name):
        return Deferred(Attr(self, name))

    def __iter__(self):
        raise TypeError(f"{self.__class__.__name__!r} object is not iterable")

    def __bool__(self):
        raise TypeError(
            f"The truth value of {self.__class__.__name__} objects is not defined"
        )

    def __getitem__(self, name):
        return Deferred(Item(self, name))

    def __call__(self, *args, **kwargs):
        return Deferred(Call(self, *args, **kwargs))

    def __invert__(self) -> Deferred:
        return Deferred(Unop(operator.invert, self))

    def __neg__(self) -> Deferred:
        return Deferred(Unop(operator.neg, self))

    def __pos__(self) -> Deferred:
        return Deferred(Unop(operator.pos, self))

    def __add__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.add, self, other))

    def __radd__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.add, other, self))

    def __sub__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.sub, self, other))

    def __rsub__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.sub, other, self))

    def __mul__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.mul, self, other))

    def __rmul__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.mul, other, self))

    def __truediv__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.truediv, self, other))

    def __rtruediv__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.truediv, other, self))

    def __floordiv__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.floordiv, self, other))

    def __rfloordiv__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.floordiv, other, self))

    def __pow__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.pow, self, other))

    def __rpow__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.pow, other, self))

    def __mod__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.mod, self, other))

    def __rmod__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.mod, other, self))

    def __rshift__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.rshift, self, other))

    def __rrshift__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.rshift, other, self))

    def __lshift__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.lshift, self, other))

    def __rlshift__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.lshift, other, self))

    def __eq__(self, other: Any) -> Deferred:  # type: ignore
        return Deferred(Binop(operator.eq, self, other))

    def __ne__(self, other: Any) -> Deferred:  # type: ignore
        return Deferred(Binop(operator.ne, self, other))

    def __lt__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.lt, self, other))

    def __le__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.le, self, other))

    def __gt__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.gt, self, other))

    def __ge__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.ge, self, other))

    def __and__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.and_, self, other))

    def __rand__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.and_, other, self))

    def __or__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.or_, self, other))

    def __ror__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.or_, other, self))

    def __xor__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.xor, self, other))

    def __rxor__(self, other: Any) -> Deferred:
        return Deferred(Binop(operator.xor, other, self))


@cython.cclass
class Builder:
    @staticmethod
    def __coerce__(value) -> Builder:
        if isinstance(value, Builder):
            return value
        elif isinstance(value, Deferred):
            return cython.cast(Deferred, value)._builder
        else:
            raise ValueError(f"Cannot coerce {type(value).__name__!r} to Builder")

    def apply(self, ctx: Context):
        return self.build(ctx)

    @cython.cfunc
    def build(self, ctx: Context): ...

    def __eq__(self, other: Any) -> bool:
        return type(self) is type(other) and self.equals(other)

    def __hash__(self):
        return self._hash()

    def __repr__(self):
        raise NotImplementedError(f"{self.__class__.__name__} is not reprable")


def _deferred_repr(obj):
    try:
        return obj.__deferred_repr__()
    except (AttributeError, TypeError):
        if callable(obj) and hasattr(obj, "__name__"):
            return obj.__name__
        else:
            return repr(obj)


@cython.final
@cython.cclass
class Func(Builder):
    """Construct a value by calling a function.

    The function is called with two positional arguments:
    1. the value being matched
    2. the context dictionary

    The function must return the constructed value.

    Parameters
    ----------
    func
        The function to apply.
    """

    func: Any

    def __init__(self, func: Any):
        self.func = func

    def __repr__(self):
        return _deferred_repr(self.func)

    def _hash(self):
        return hash((self.__class__, self.func))

    def equals(self, other: Func) -> bool:
        return self.func == other.func

    @cython.cfunc
    def build(self, ctx: Context):
        return self.func(**ctx)


@cython.final
@cython.cclass
class Just(Builder):
    """Construct exactly the given value.

    Parameters
    ----------
    value
        The value to return when the deferred is called.
    """

    value: Any

    def __init__(self, value: Any):
        if isinstance(value, Just):
            self.value = cython.cast(Just, value).value
        elif isinstance(value, (Builder, Deferred)):
            raise TypeError(f"`{value}` cannot be used as a Just value")
        else:
            self.value = value

    def __repr__(self):
        return _deferred_repr(self.value)

    def _hash(self):
        return hash((self.__class__, PseudoHashable(self.value)))

    def equals(self, other: Just) -> bool:
        return self.value == other.value

    @cython.cfunc
    def build(self, ctx: Context):
        return self.value


@cython.cclass
class Var(Builder):
    """Retrieve a value from the context.

    Parameters
    ----------
    name
        The key to retrieve from the state.
    """

    name = cython.declare(str, visibility="readonly")

    def __init__(self, name: str):
        self.name = name

    def __repr__(self):
        return f"${self.name}"

    def _hash(self):
        return hash((self.__class__, self.name))

    def equals(self, other: Var) -> bool:
        return self.name == other.name

    @cython.cfunc
    def build(self, ctx: Context):
        return ctx[self.name]


def Call(func, *args, **kwargs) -> Builder:
    """Call a function with the given arguments.

    Parameters
    ----------
    func
        The function to call.
    args
        The positional arguments to pass to the function.
    kwargs
        The keyword arguments to pass to the function.
    """
    if kwargs:
        return CallN(func, *args, **kwargs)
    elif len(args) == 0:
        return Call0(func)
    elif len(args) == 1:
        return Call1(func, *args)
    elif len(args) == 2:
        return Call2(func, *args)
    elif len(args) == 3:
        return Call3(func, *args)
    else:
        return CallN(func, *args)


@cython.final
@cython.cclass
class Call0(Builder):
    """Pattern that calls a function with no arguments.

    Parameters
    ----------
    func
        The function to call.
    """

    func: Builder

    def __init__(self, func):
        self.func = builder(func)

    def __repr__(self):
        return f"{self.func!r}()"

    def _hash(self):
        return hash((self.__class__, self.func))

    def equals(self, other: Call0) -> bool:
        return self.func == other.func

    @cython.cfunc
    def build(self, ctx: Context):
        func = self.func.build(ctx)
        return func()


@cython.final
@cython.cclass
class Call1(Builder):
    """Pattern that calls a function with one argument.

    Parameters
    ----------
    func
        The function to call.
    arg
        The argument to pass to the function.
    """

    func: Builder
    arg: Builder

    def __init__(self, func, arg):
        self.func = builder(func)
        self.arg = builder(arg)

    def __repr__(self):
        return f"{self.func!r}({self.arg!r})"

    def _hash(self):
        return hash((self.__class__, self.func, self.arg))

    def equals(self, other: Call1) -> bool:
        return self.func == other.func and self.arg == other.arg

    @cython.cfunc
    def build(self, ctx: Context):
        func = self.func.build(ctx)
        arg = self.arg.build(ctx)
        return func(arg)


@cython.final
@cython.cclass
class Call2(Builder):
    """Pattern that calls a function with two arguments.

    Parameters
    ----------
    func
        The function to call.
    arg1
        The first argument to pass to the function.
    arg2
        The second argument to pass to the function.
    """

    func: Builder
    arg1: Builder
    arg2: Builder

    def __init__(self, func, arg1, arg2):
        self.func = builder(func)
        self.arg1 = builder(arg1)
        self.arg2 = builder(arg2)

    def __repr__(self):
        return f"{self.func!r}({self.arg1!r}, {self.arg2!r})"

    def _hash(self):
        return hash((self.__class__, self.func, self.arg1, self.arg2))

    def equals(self, other: Call2) -> bool:
        return (
            self.func == other.func
            and self.arg1 == other.arg1
            and self.arg2 == other.arg2
        )

    @cython.cfunc
    def build(self, ctx: Context):
        func = self.func.build(ctx)
        arg1 = self.arg1.build(ctx)
        arg2 = self.arg2.build(ctx)
        return func(arg1, arg2)


@cython.final
@cython.cclass
class Call3(Builder):
    """Pattern that calls a function with three arguments.

    Parameters
    ----------
    func
        The function to call.
    arg1
        The first argument to pass to the function.
    arg2
        The second argument to pass to the function.
    arg3
        The third argument to pass to the function.
    """

    func: Builder
    arg1: Builder
    arg2: Builder
    arg3: Builder

    def __init__(self, func, arg1, arg2, arg3):
        self.func = builder(func)
        self.arg1 = builder(arg1)
        self.arg2 = builder(arg2)
        self.arg3 = builder(arg3)

    def __repr__(self):
        return f"{self.func!r}({self.arg1!r}, {self.arg2!r}, {self.arg3!r})"

    def _hash(self):
        return hash((self.__class__, self.func, self.arg1, self.arg2, self.arg3))

    def equals(self, other: Call3) -> bool:
        return (
            self.func == other.func
            and self.arg1 == other.arg1
            and self.arg2 == other.arg2
            and self.arg3 == other.arg3
        )

    @cython.cfunc
    def build(self, ctx: Context):
        func = self.func.build(ctx)
        arg1 = self.arg1.build(ctx)
        arg2 = self.arg2.build(ctx)
        arg3 = self.arg3.build(ctx)
        return func(arg1, arg2, arg3)


@cython.final
@cython.cclass
class CallN(Builder):
    """Pattern that calls a function with the given arguments.

    Both positional and keyword arguments are coerced into patterns.

    Parameters
    ----------
    func
        The function to call.
    args
        The positional argument patterns.
    kwargs
        The keyword argument patterns.
    """

    func: Builder
    args: tuple[Builder, ...]
    kwargs: dict[str, Builder]

    def __init__(self, func, *args, **kwargs):
        self.func = builder(func)
        self.args = tuple(builder(arg) for arg in args)
        self.kwargs = {k: builder(v) for k, v in kwargs.items()}

    def __repr__(self):
        args = ", ".join(map(repr, self.args))
        kwargs = ", ".join(f"{k}={v!r}" for k, v in self.kwargs.items())
        if self.args and self.kwargs:
            return f"{self.func!r}({args}, {kwargs})"
        elif self.args:
            return f"{self.func!r}({args})"
        elif self.kwargs:
            return f"{self.func!r}({kwargs})"
        else:
            return f"{self.func!r}()"

    def _hash(self):
        return hash((self.__class__, self.func, self.args, PseudoHashable(self.kwargs)))

    def equals(self, other: CallN) -> bool:
        return (
            self.func == other.func
            and self.args == other.args
            and self.kwargs == other.kwargs
        )

    @cython.cfunc
    def build(self, ctx: Context):
        arg: Builder
        args: list[Any] = []
        kwargs: dict[str, Any] = {}

        func = self.func.build(ctx)
        for arg in self.args:
            args.append(arg.build(ctx))
        for key, arg in self.kwargs.items():
            kwargs[key] = arg.build(ctx)

        return func(*args, **kwargs)


_operator_symbols = {
    operator.add: "+",
    operator.sub: "-",
    operator.mul: "*",
    operator.truediv: "/",
    operator.floordiv: "//",
    operator.pow: "**",
    operator.mod: "%",
    operator.eq: "==",
    operator.ne: "!=",
    operator.lt: "<",
    operator.le: "<=",
    operator.gt: ">",
    operator.ge: ">=",
    operator.and_: "&",
    operator.or_: "|",
    operator.xor: "^",
    operator.rshift: ">>",
    operator.lshift: "<<",
    operator.inv: "~",
    operator.neg: "-",
    operator.invert: "~",
}


@cython.final
@cython.cclass
class Unop(Builder):
    """Pattern that applies a unary operator to a value.

    Parameters
    ----------
    op
        The unary operator to apply.
    arg
        The argument to apply the operator to.
    """

    op: Any
    arg: Builder

    def __init__(self, op: Any, arg: Any):
        self.op = op
        self.arg = builder(arg)

    def __repr__(self):
        symbol = _operator_symbols[self.op]
        return f"{symbol}{self.arg!r}"

    def _hash(self):
        return hash((self.__class__, self.op, self.arg))

    def equals(self, other: Unop) -> bool:
        return self.op == other.op and self.arg == other.arg

    @cython.cfunc
    def build(self, ctx: Context):
        arg = self.arg.build(ctx)
        return self.op(arg)


@cython.final
@cython.cclass
class Binop(Builder):
    """Pattern that applies a binary operator to two values.

    Parameters
    ----------
    op
        The binary operator to apply.
    arg1
        The left-hand side argument.
    arg2
        The right-hand side argument.
    """

    op: Any
    arg1: Builder
    arg2: Builder

    def __init__(self, op: Any, arg1: Any, arg2: Any):
        self.op = op
        self.arg1 = builder(arg1)
        self.arg2 = builder(arg2)

    def __repr__(self):
        symbol = _operator_symbols[self.op]
        return f"({self.arg1!r} {symbol} {self.arg2!r})"

    def _hash(self):
        return hash((self.__class__, self.op, self.arg1, self.arg2))

    def equals(self, other: Binop) -> bool:
        return (
            self.op == other.op and self.arg1 == other.arg1 and self.arg2 == other.arg2
        )

    @cython.cfunc
    def build(self, ctx: Context):
        arg1 = self.arg1.build(ctx)
        arg2 = self.arg2.build(ctx)
        return self.op(arg1, arg2)


@cython.final
@cython.cclass
class Item(Builder):
    """Pattern that retrieves an item from a container.

    Parameters
    ----------
    container
        The container to retrieve the item from.
    key
        The key to retrieve from the container.
    """

    obj: Builder
    key: Builder

    def __init__(self, obj, key):
        self.obj = builder(obj)
        self.key = builder(key)

    def __repr__(self):
        return f"{self.obj!r}[{self.key!r}]"

    def _hash(self):
        return hash((self.__class__, self.obj, self.key))

    def equals(self, other: Item) -> bool:
        return self.obj == other.obj and self.key == other.key

    @cython.cfunc
    def build(self, ctx: Context):
        obj = self.obj.build(ctx)
        key = self.key.build(ctx)
        return obj[key]


@cython.final
@cython.cclass
class Attr(Builder):
    """Pattern that retrieves an attribute from an object.

    Parameters
    ----------
    obj
        The object to retrieve the attribute from.
    attr
        The attribute to retrieve from the object.
    """

    obj: Builder
    attr: str

    def __init__(self, obj: Any, attr: str):
        self.obj = builder(obj)
        self.attr = attr

    def __repr__(self):
        return f"{self.obj!r}.{self.attr}"

    def _hash(self):
        return hash((self.__class__, self.obj, self.attr))

    def equals(self, other: Attr) -> bool:
        return self.obj == other.obj and self.attr == other.attr

    @cython.cfunc
    def build(self, ctx: Context):
        obj = self.obj.build(ctx)
        return getattr(obj, self.attr)


@cython.final
@cython.cclass
class Seq(Builder):
    """Pattern that constructs a sequence from the given items.

    Parameters
    ----------
    items
        The items to construct the sequence from.
    """

    type_: Any
    items: tuple[Builder, ...]

    def __init__(self, items):
        self.type_ = type(items)
        self.items = tuple(builder(item) for item in items)

    def __repr__(self):
        elems = ", ".join(map(repr, self.items))
        if self.type_ is tuple:
            return f"({elems})"
        elif self.type_ is list:
            return f"[{elems}]"
        else:
            return f"{self.type_.__name__}({elems})"

    def _hash(self):
        return hash((self.__class__, self.type_, self.items))

    def equals(self, other: Seq) -> bool:
        return self.type_ == other.type_ and self.items == other.items

    @cython.cfunc
    def build(self, ctx: Context):
        item: Builder
        result: list[Any] = []
        for item in self.items:
            result.append(item.build(ctx))
        return self.type_(result)


@cython.final
@cython.cclass
class Map(Builder):
    """Pattern that constructs a mapping from the given items.

    Parameters
    ----------
    items
        The items to construct the mapping from.
    """

    type_: Any
    items: dict[Any, Builder]

    def __init__(self, items):
        self.type_ = type(items)
        self.items = {k: builder(v) for k, v in items.items()}

    def __repr__(self):
        items = ", ".join(f"{k!r}: {v!r}" for k, v in self.items.items())
        if self.type_ is dict:
            return f"{{{items}}}"
        else:
            return f"{self.type_.__name__}({{{items}}})"

    def _hash(self):
        return hash((self.__class__, self.type_, PseudoHashable(self.items)))

    def equals(self, other: Map) -> bool:
        return self.type_ == other.type_ and self.items == other.items

    @cython.cfunc
    def build(self, ctx: Context):
        k: Any
        v: Builder
        result: dict = {}
        for k, v in self.items.items():
            result[k] = v.build(ctx)
        return self.type_(result)


@cython.ccall
def builder(obj, allow_custom=False) -> Builder:
    if isinstance(obj, Deferred):
        return cython.cast(Deferred, obj)._builder
    elif isinstance(obj, Builder):
        return obj
    elif isinstance(obj, collections.abc.Mapping):
        # allow nesting deferred patterns in dicts
        return Map(obj)
    elif isinstance(obj, collections.abc.Sequence):
        # allow nesting deferred patterns in tuples/lists
        if isinstance(obj, (str, bytes)):
            return Just(obj)
        else:
            return Seq(obj)
    elif callable(obj) and allow_custom:
        # the object is used as a custom builder function
        return Func(obj)
    else:
        # the object is used as a constant value
        return Just(obj)


def deferred(obj, allow_custom=False) -> Deferred:
    return Deferred(builder(obj, allow_custom))


def resolve(obj, **context):
    bldr: Builder = builder(obj)
    return bldr.build(context)


def _contains_deferred(obj: Any) -> bool:
    if isinstance(obj, (Builder, Deferred)):
        return True
    elif (typ := type(obj)) in (tuple, list, set):
        return any(_contains_deferred(o) for o in obj)
    elif typ is dict:
        return any(_contains_deferred(o) for o in obj.values())
    return False


def deferrable(func=None, *, repr=None):
    """Wrap a top-level expr function to support deferred arguments.

    When a deferrable function is called, the args & kwargs are traversed to
    look for `Deferred` values (through builtin collections like
    `list`/`tuple`/`set`/`dict`). If any `Deferred` arguments are found, then
    the result is also `Deferred`. Otherwise the function is called directly.

    Parameters
    ----------
    func
        A callable to make deferrable
    repr
        An optional fixed string to use when repr-ing the deferred expression,
        instead of the usual. This is useful for complex deferred expressions
        where the arguments don't necessarily make sense to be user facing
        in the repr.

    """

    def wrapper(func):
        # Parse the signature of func so we can validate deferred calls eagerly,
        # erroring for invalid/missing arguments at call time not resolve time.
        sig = inspect.signature(func)

        @functools.wraps(func)
        def inner(*args, **kwargs):
            if _contains_deferred((args, kwargs)):
                # Try to bind the arguments now, raising a nice error
                # immediately if the function was called incorrectly
                sig.bind(*args, **kwargs)
                builder = Call(func, *args, **kwargs)
                return Deferred(builder, repr=repr)
            return func(*args, **kwargs)

        return inner  # type: ignore

    return wrapper if func is None else wrapper(func)



from __future__ import annotations

import itertools
import sys
import typing
from collections.abc import Hashable, Mapping, Sequence, Set
from typing import Any, ClassVar, ForwardRef, Optional, TypeVar

from typing_extensions import Self

_SpecialForm = type(Self)

K = TypeVar("K")
V = TypeVar("V")

get_type_args = typing.get_args
get_type_origin = typing.get_origin
_get_type_hints = typing.get_type_hints


class FakeType:
    def __init__(self, module: str, annotations: dict[str, Any]):
        self.__module__ = module
        self.__annotations__ = annotations


def get_type_hints(
    obj: Any,
    globalns=None,
    localns=None,
    module: Optional[str] = None,
    include_extras: bool = True,
    include_properties: bool = False,
) -> dict[str, Any]:
    """Get type hints for a callable or class.

    Extension of typing.get_type_hints that supports getting type hints for
    class properties.

    Parameters
    ----------
    obj
        Callable or class to get type hints for.
    globalns
        Global namespace to use for resolving type variables.
    localns
        Local namespace to use for resolving type variables.
    module
        Module name to use for resolving globalns if not provided.
    include_extras
        Whether to include extra type hints such as Annotated.
    include_properties
        Whether to include type hints for class properties.

    Returns
    -------
    Mapping of parameter or attribute name to type hint.
    """
    if isinstance(obj, dict):
        if module is None:
            raise TypeError("`module` must be provided to evaluate type hints")

        # turn string annotations into ForwardRef objects with the
        # is_class flag set, so ClassVar annotations won't raise
        # on Python 3.10
        obj = {
            k: ForwardRef(v, is_class=True) if isinstance(v, str) else v
            for k, v in obj.items()
        }
        obj = FakeType(module=module, annotations=obj)
        mod = sys.modules.get(module, None)
        globalns = getattr(mod, "__dict__", None)

    hints = _get_type_hints(
        obj, globalns=globalns, localns=localns, include_extras=include_extras
    )
    if include_properties:
        for name in dir(obj):
            attr = getattr(obj, name)
            if isinstance(attr, property):
                annots = _get_type_hints(
                    attr.fget,
                    globalns=globalns,
                    localns=localns,
                    include_extras=include_extras,
                )
                if return_hint := annots.get("return"):
                    hints[name] = return_hint

    return hints


def get_type_params(typ: Any) -> dict[TypeVar, type]:
    """Get type parameters for a generic class.

    Parameters
    ----------
    typ
        Generic class to get type parameters for.

    Returns
    -------
    Mapping of type parameter name to type.

    Examples
    --------
    >>> from typing import Dict, List
    >>> class MyList(List[T]): ...
    >>> get_type_params(MyList[int])
    {'T': <class 'int'>}
    >>> class MyDict(Dict[T, U]): ...
    >>> get_type_params(MyDict[int, str])
    {'T': <class 'int'>, 'U': <class 'str'>}
    """
    args: tuple = get_type_args(typ)
    origin: Any = get_type_origin(typ)  # or typ
    params: tuple = getattr(origin, "__parameters__", ())

    result: dict[str, type] = {}
    for param, arg in zip(params, args):
        result[param.__name__] = arg

    return result


def get_type_boundvars(typ: Any) -> dict[TypeVar, tuple[str, type]]:
    """Get the attribute names and concrete types of a generic class.

    Parameters
    ----------
    typ
        Generic class to get type variables for.
    only_covariant
        Whether to only include covariant type variables.

    Returns
    -------
    Mapping of type variable to attribute name and type.

    Examples
    --------
    >>> from typing import Generic
    >>> class MyStruct(Generic[T, U]):
    ...     a: T
    ...     b: U
    >>> get_bound_typevars(MyStruct[int, str])
    {~T: ('a', <class 'int'>), ~U: ('b', <class 'str'>)}
    >>>
    >>> class MyStruct(Generic[T, U]):
    ...     a: T
    ...
    ...     @property
    ...     def myprop(self) -> U: ...
    >>> get_bound_typevars(MyStruct[float, bytes])
    {~T: ('a', <class 'float'>), ~U: ('myprop', <class 'bytes'>)}
    """
    origin = get_type_origin(typ) or typ
    hints = get_type_hints(origin, include_properties=True)
    params = get_type_params(typ)

    result = {}
    for attr, hint in hints.items():
        if isinstance(hint, TypeVar) and hint.__name__ in params:
            if not hint.__covariant__:
                raise TypeError(
                    f"Typevar {hint} is not covariant and currently only "
                    "covariant typevars are supported"
                )
            result[attr] = params.pop(hint.__name__)

    if params:
        raise ValueError(
            f"Unable to deduce corresponding attributes for type parameters of {typ}.\n"
            f"Missing attributes with typehints for the following type variables: {params}.\n"
            f"Available type hints: {hints}"
        )

    return result


_hint_types = (type, TypeVar, _SpecialForm)


def is_typehint(obj: Any) -> bool:
    return isinstance(obj, _hint_types) or get_type_origin(obj)


class FrozenDict(dict[K, V], Hashable):
    __slots__ = ("__precomputed_hash__",)
    __precomputed_hash__: int

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        hashable = frozenset(self.items())
        object.__setattr__(self, "__precomputed_hash__", hash(hashable))

    def __hash__(self) -> int:
        return self.__precomputed_hash__

    def __setitem__(self, key: K, value: V) -> None:
        raise TypeError(
            f"'{self.__class__.__name__}' object does not support item assignment"
        )

    def __setattr__(self, name: str, _: Any) -> None:
        raise TypeError(f"Attribute {name!r} cannot be assigned to frozendict")

    def __reduce__(self) -> tuple:
        return (self.__class__, (dict(self),))


frozendict = FrozenDict


class RewindableIterator:
    """Iterator that can be rewound to a checkpoint.

    Examples
    --------
    >>> it = RewindableIterator(range(5))
    >>> next(it)
    0
    >>> next(it)
    1
    >>> it.checkpoint()
    >>> next(it)
    2
    >>> next(it)
    3
    >>> it.rewind()
    >>> next(it)
    2
    >>> next(it)
    3
    >>> next(it)
    4

    """

    __slots__ = ("_iterator", "_checkpoint")

    def __init__(self, iterable):
        self._iterator = iter(iterable)
        self._checkpoint = None

    def __next__(self):
        return next(self._iterator)

    def rewind(self):
        """Rewind the iterator to the last checkpoint."""
        if self._checkpoint is None:
            raise ValueError("No checkpoint to rewind to.")
        self._iterator, self._checkpoint = itertools.tee(self._checkpoint)

    def checkpoint(self):
        """Create a checkpoint of the current iterator state."""
        self._iterator, self._checkpoint = itertools.tee(self._iterator)


class PseudoHashable:
    """A wrapper that provides a best effort precomputed hash."""

    __slots__ = ("obj", "hash")

    def __new__(cls, obj):
        if isinstance(obj, Hashable):
            return obj
        else:
            return super().__new__(cls)

    def __init__(self, obj):
        if isinstance(obj, Sequence):
            hashable_obj = tuple(obj)
        elif isinstance(obj, Mapping):
            hashable_obj = tuple(obj.items())
        elif isinstance(obj, Set):
            hashable_obj = frozenset(obj)
        else:
            hashable_obj = id(obj)

        self.obj = obj
        self.hash = hash((type(obj), hashable_obj))

    def __hash__(self):
        return self.hash

    def __eq__(self, other):
        if isinstance(other, PseudoHashable):
            return self.obj == other.obj
        else:
            return NotImplemented


# def format_typehint(typ: Any) -> str:
#     if isinstance(typ, type):
#         return typ.__name__
#     elif isinstance(typ, TypeVar):
#         if typ.__bound__ is None:
#             return str(typ)
#         else:
#             return format_typehint(typ.__bound__)
#     else:
#         # remove the module name from the typehint, including generics
#         return re.sub(r"(\w+\.)+", "", str(typ))
